@@ -8,7 +8,11 @@ var pathPrinter = {
 }
 
 var textPrinter = {
+	beginSpan: function (f,m,wmode, bidi, dir, lang) {
+		print("beginSpan",f,m,wmode,bidi,dir,repr(lang));
+	},
 	showGlyph: function (f,m,g,u,v,b) { print("glyph",f,m,g,u,v,b) },
+	endSpan: function () { print("endSpan"); }
 }
 
 var traceDevice = {
@@ -87,6 +91,33 @@ var traceDevice = {
 	endTile: function () {
 		print("endTile")
 	},
+	beginLayer: function (name) {
+		print("beginLayer", name)
+	},
+	endLayer: function () {
+		print("endLayer")
+	},
+	beginStructure: function (structure, raw, uid) {
+		print("beginStructure", structure, raw, uiw)
+	},
+	endStructure: function () {
+		print("endStructure")
+	},
+	beginMetatext: function (meta, metatext) {
+		print("beginMetatext", meta, metatext)
+	},
+	endMetatext: function () {
+		print("endMetatext")
+	},
+
+	renderFlags: function (set, clear) {
+		print("renderFlags", set, clear)
+	},
+	setDefaultColorSpaces: function (colorSpaces) {
+		print("setDefaultColorSpaces", colorSpaces.getDefaultGray(),
+		colorSpaces.getDefaultRGB(), colorSpaces.getDefaultCMYK(),
+		colorSpaces.getOutputIntent())
+	},
 
 	close: function () {
 		print("close")
@@ -96,7 +127,7 @@ var traceDevice = {
 if (scriptArgs.length != 2)
 	print("usage: mutool run trace-device.js document.pdf pageNumber")
 else {
-	var doc = new Document(scriptArgs[0]);
+	var doc = Document.openDocument(scriptArgs[0]);
 	var page = doc.loadPage(parseInt(scriptArgs[1])-1);
-	page.run(traceDevice, Identity);
+	page.run(traceDevice, Matrix.identity);
 }
