@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2023 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 package example;
 
@@ -169,6 +169,18 @@ public class TraceDevice extends Device implements PathWalker, TextWalker
 		System.out.println("endTile");
 	}
 
+	public void renderFlags(int set, int clear) {
+		System.out.println("renderFlags set=" + set + " clear=" + clear);
+	}
+
+	public void setDefaultColorSpaces(DefaultColorSpaces dcs) {
+		System.out.println("setDefaultColorSpaces" +
+			" gray=" + dcs.getDefaultGray() +
+			" rgb=" + dcs.getDefaultRGB() +
+			" cmyk=" + dcs.getDefaultCMYK() +
+			" outputIntent=" + dcs.getOutputIntent());
+	}
+
 	public void beginLayer(String name) {
 		System.out.println("beginLayer");
 	}
@@ -177,6 +189,25 @@ public class TraceDevice extends Device implements PathWalker, TextWalker
 		System.out.println("endLayer");
 	}
 
+	public void beginStructure(int standard, String raw, int uid) {
+		System.out.println("beginStructure standard=" + standard +
+			" raw=" + raw +
+			" uid=" + uid);
+	}
+
+	public void endStructure() {
+		System.out.println("endStructure");
+	}
+
+	public void beginMetatext(int meta, String text) {
+		System.out.println("beginMetatext type=" + meta +
+			" text=" + text);
+	}
+
+	public void endMetatext() {
+		System.out.println("endMetatext");
+	}
+
 	public static void main(String[] args) {
 		Document doc = Document.openDocument("pdfref17.pdf");
 		Page page = doc.loadPage(1144);
