@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2024 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -134,15 +134,14 @@ FUN(DisplayList_toStructuredText)(JNIEnv *env, jobject self, jstring joptions)
 	return to_StructuredText_safe_own(ctx, env, text);
 }
 
-JNIEXPORT jobject JNICALL
+JNIEXPORT jobjectArray JNICALL
 FUN(DisplayList_search)(JNIEnv *env, jobject self, jstring jneedle)
 {
 	fz_context *ctx = get_context(env);
 	fz_display_list *list = from_DisplayList(env, self);
-	fz_quad hits[500];
-	int marks[500];
 	const char *needle = NULL;
-	int n = 0;
+	search_state state = { env, NULL, 0 };
+	jobject jsample = NULL;
 
 	if (!ctx || !list) return NULL;
 	if (!jneedle) jni_throw_arg(env, "needle must not be null");
@@ -150,14 +149,23 @@ FUN(DisplayList_search)(JNIEnv *env, jobject self, jstring jneedle)
 	needle = (*env)->GetStringUTFChars(env, jneedle, NULL);
 	if (!needle) return NULL;
 
+	state.hits = (*env)->NewObject(env, cls_ArrayList, mid_ArrayList_init);
+	if (!state.hits || (*env)->ExceptionCheck(env)) return NULL;
+
 	fz_try(ctx)
-		n = fz_search_display_list(ctx, list, needle, marks, hits, nelem(hits));
+		fz_search_display_list_cb(ctx, list, needle, hit_callback, &state);
 	fz_always(ctx)
+	{
 		(*env)->ReleaseStringUTFChars(env, jneedle, needle);
+	}
 	fz_catch(ctx)
 		jni_rethrow(env, ctx);
 
-	return to_SearchHits_safe(ctx, env, marks, hits, n);
+	if (state.error)
+		return NULL;
+
+	jsample = (*env)->NewObjectArray(env, 0, cls_ArrayOfQuad, NULL);
+	return (*env)->CallObjectMethod(env, state.hits, mid_ArrayList_toArray, jsample);
 }
 
 JNIEXPORT jobject JNICALL
