@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2024 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -72,12 +72,12 @@ FUN(PDFPage_update)(JNIEnv *env, jobject self)
 }
 
 JNIEXPORT jboolean JNICALL
-FUN(PDFPage_applyRedactions)(JNIEnv *env, jobject self, jboolean blackBoxes, jint imageMethod)
+FUN(PDFPage_applyRedactions)(JNIEnv *env, jobject self, jboolean blackBoxes, jint imageMethod, jint lineArt, jint text)
 {
 	fz_context *ctx = get_context(env);
 	pdf_page *page = from_PDFPage(env, self);
 	jboolean redacted = JNI_FALSE;
-	pdf_redact_options opts = { blackBoxes, imageMethod };
+	pdf_redact_options opts = { blackBoxes, imageMethod, lineArt, text };
 
 	if (!ctx || !page) return JNI_FALSE;
 
@@ -253,3 +253,49 @@ FUN(PDFPage_getObject)(JNIEnv *env, jobject self)
 
 	return to_PDFObject_safe_own(ctx, env, pdf_keep_obj(ctx, page->obj));
 }
+
+JNIEXPORT void JNICALL
+FUN(PDFPage_setPageBox)(JNIEnv *env, jobject self, jint box, jobject jrect)
+{
+	fz_context *ctx = get_context(env);
+	pdf_page *page = from_PDFPage(env, self);
+	fz_rect rect = from_Rect(env, jrect);
+
+	if (!ctx || !page)
+		return;
+
+	fz_try(ctx)
+		pdf_set_page_box(ctx, page, box, rect);
+	fz_catch(ctx)
+		jni_rethrow_void(env, ctx);
+}
+
+JNIEXPORT jint JNICALL
+FUN(PDFPage_countAssociatedFiles)(JNIEnv *env, jobject self)
+{
+	fz_context *ctx = get_context(env);
+	pdf_page *page = from_PDFPage(env, self);
+	int n;
+
+	fz_try(ctx)
+		n = pdf_count_page_associated_files(ctx, page);
+	fz_catch(ctx)
+		jni_rethrow(env, ctx);
+
+	return n;
+}
+
+JNIEXPORT jobject JNICALL
+FUN(PDFPage_associatedFile)(JNIEnv *env, jobject self, jint idx)
+{
+	fz_context *ctx = get_context(env);
+	pdf_page *page = from_PDFPage(env, self);
+	pdf_obj *af;
+
+	fz_try(ctx)
+		af = pdf_page_associated_file(ctx, page, idx);
+	fz_catch(ctx)
+		jni_rethrow(env, ctx);
+
+	return to_PDFObject_safe_own(ctx, env, af);
+}
