@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2023 Artifex Software, Inc.
+// Copyright (C) 2004-2024 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -83,7 +83,7 @@ abstract public class Device
 	abstract public void setDefaultColorSpaces(DefaultColorSpaces dcs);
 	abstract public void beginLayer(String name);
 	abstract public void endLayer();
-	abstract public void beginStructure(int standard, String raw, int uid);
+	abstract public void beginStructure(int standard, String raw, int idx);
 	abstract public void endStructure();
 	abstract public void beginMetatext(int meta, String text);
 	abstract public void endMetatext();
