@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2023 Artifex Software, Inc.
+// Copyright (C) 2004-2024 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -123,26 +123,28 @@ public class PDFAnnotation
 	public static final int IS_TOGGLE_NO_VIEW = 1 << (9-1);
 	public static final int IS_LOCKED_CONTENTS = 1 << (10-1);
 
+	public static final int IT_DEFAULT = 0;
+	public static final int IT_FREETEXT_CALLOUT = 1;
+	public static final int IT_FREETEXT_TYPEWRITER = 2;
+	public static final int IT_LINE_ARROW = 3;
+	public static final int IT_LINE_DIMENSION = 4;
+	public static final int IT_POLYLINE_DIMENSION = 5;
+	public static final int IT_POLYGON_CLOUD = 6;
+	public static final int IT_POLYGON_DIMENSION = 7;
+	public static final int IT_STAMP_IMAGE = 8;
+	public static final int IT_STAMP_SNAPSHOT = 9;
+	public static final int IT_UNKNOWN = 255;
+
 	public native int getType();
 	public native int getFlags();
 	public native void setFlags(int flags);
 	public native String getContents();
 	public native void setContents(String contents);
-	public native boolean hasRect();
-	public native Rect getRect();
-	public native void setRect(Rect rect);
-	public native float getBorder();
-	public native void setBorder(float width);
 	public native float[] getColor();
 	public native void setColor(float[] color);
-	public native boolean hasInteriorColor();
-	public native float[] getInteriorColor();
-	public native void setInteriorColor(float[] color);
 	public native float getOpacity();
 	public native void setOpacity(float opacity);
-	public native boolean hasAuthor();
-	public native String getAuthor();
-	public native void setAuthor(String author);
+
 	protected native long getCreationDateNative();
 	protected native void setCreationDate(long time);
 	protected native long getModificationDateNative();
@@ -160,6 +162,18 @@ public class PDFAnnotation
 		setModificationDate(date.getTime());
 	}
 
+	public native boolean hasRect();
+	public native Rect getRect();
+	public native void setRect(Rect rect);
+
+	public native boolean hasInteriorColor();
+	public native float[] getInteriorColor();
+	public native void setInteriorColor(float[] color);
+
+	public native boolean hasAuthor();
+	public native String getAuthor();
+	public native void setAuthor(String author);
+
 	public native boolean hasLineEndingStyles();
 	public native int[] getLineEndingStyles();
 	public native void setLineEndingStyles(int startStyle, int endStyle);
@@ -263,13 +277,64 @@ public class PDFAnnotation
 		return list;
 	}
 
+	public native boolean hasCallout();
+	public native int getCalloutStyle();
+	public native void setCalloutStyle(int s);
+	public native Point getCalloutPoint();
+	public native void setCalloutPoint(Point p);
+	public native Point[] getCalloutLine();
+	public native void setCalloutLineNative(int n, Point a, Point b, Point c);
+	public void setCalloutLine() {
+		setCalloutLineNative(0, null, null, null);
+	}
+	public void setCalloutLine(Point[] line) {
+		if (line.length == 0)
+			setCalloutLineNative(0, null, null, null);
+		else if (line.length == 2)
+			setCalloutLineNative(2, line[0], line[1], null);
+		else if (line.length == 3)
+			setCalloutLineNative(3, line[0], line[1], line[2]);
+		else
+			throw new IllegalArgumentException("Callout Line must have 0, 2, or 3 points.");
+	}
+
 	public native boolean hasIcon();
 	public native String getIcon();
 	public native void setIcon(String icon);
+
+	public native boolean hasPopup();
+	public native Rect getPopup();
+	public native void setPopup(Rect rect);
+
 	public native boolean hasOpen();
-	public native boolean isOpen();
+	public native boolean getIsOpen();
 	public native void setIsOpen(boolean open);
 
+	public native boolean hasLine();
+	public native Point[] getLine();
+	public native void setLine(Point a, Point b);
+	public native float getLineLeader();
+	public native void setLineLeader(float length);
+	public native float getLineLeaderExtension();
+	public native void setLineLeaderExtension(float extension);
+	public native float getLineLeaderOffset();
+	public native void setLineLeaderOffset(float offset);
+	public native boolean getLineCaption();
+	public native void setLineCaption(boolean caption);
+	public native Point getLineCaptionOffset();
+	public native void setLineCaptionOffset(Point offset);
+
+	public native boolean hasFilespec();
+	public native void setFilespec(PDFObject fs);
+	public native PDFObject getFilespec();
+	public boolean hasFileSpecification() { return hasFilespec(); }
+	public void setFileSpecification(PDFObject fs) { setFilespec(fs); }
+	public PDFObject getFileSpecification() { return getFilespec(); }
+
+	public native boolean hasIntent();
+	public native int getIntent();
+	public native void setIntent(int intent);
+
 	public native void eventEnter();
 	public native void eventExit();
 	public native void eventDown();
@@ -285,13 +350,10 @@ public class PDFAnnotation
 	public native int getLanguage();
 	public native void setLanguage(int lang);
 
+	public native boolean hasQuadding();
 	public native int getQuadding();
 	public native void setQuadding(int quadding);
 
-	public native boolean hasLine();
-	public native Point[] getLine();
-	public native void setLine(Point a, Point b);
-
 	public native DefaultAppearance getDefaultAppearance();
 	public native void setDefaultAppearance(String font, float size, float[] color);
 
@@ -333,12 +395,29 @@ public class PDFAnnotation
 		setNativeAppearanceImage(image);
 	}
 
-	public native boolean hasFileSpecification();
-	public native void setFileSpecification(PDFObject fs);
-	public native PDFObject getFileSpecification();
-
 	public native boolean getHiddenForEditing();
 	public native void setHiddenForEditing(boolean hidden);
 
-	public native boolean applyRedaction(boolean blackBoxes, int imageMethod);
+	public boolean applyRedaction(boolean blackBoxes)
+	{
+		return applyRedaction(blackBoxes, PDFPage.REDACT_IMAGE_PIXELS, PDFPage.REDACT_LINE_ART_NONE, PDFPage.REDACT_TEXT_REMOVE);
+	}
+
+	public boolean applyRedaction(boolean blackBoxes, int imageMethod)
+	{
+		return applyRedaction(blackBoxes, imageMethod, PDFPage.REDACT_LINE_ART_NONE, PDFPage.REDACT_TEXT_REMOVE);
+	}
+
+	public boolean applyRedaction(boolean blackBoxes, int imageMethod, int lineArt)
+	{
+		return applyRedaction(blackBoxes, imageMethod, lineArt, PDFPage.REDACT_TEXT_REMOVE);
+	}
+
+	public native boolean applyRedaction(boolean blackBoxes, int imageMethod, int lineArt, int text);
+
+	/** @deprecated use getBorderWidth instead. */
+	public float getBorder() { return getBorderWidth(); }
+
+	/** @deprecated use setBorderWidth instead. */
+	public void setBorder(float width) { setBorderWidth(width); }
 }
