@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2023 Artifex Software, Inc.
+// Copyright (C) 2004-2024 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -139,6 +139,9 @@ public class PDFObject implements Iterable<PDFObject>
 	private native void putDictionaryStringFloat(String name, float f);
 	private native void putDictionaryStringString(String name, String str);
 	private native void putDictionaryStringPDFObject(String name, PDFObject obj);
+	private native void putDictionaryStringRect(String name, Rect r);
+	private native void putDictionaryStringMatrix(String name, Matrix m);
+	private native void putDictionaryStringDate(String name, long secs);
 
 	private native void putDictionaryPDFObjectBoolean(PDFObject name, boolean b);
 	private native void putDictionaryPDFObjectInteger(PDFObject name, int i);
@@ -189,6 +192,18 @@ public class PDFObject implements Iterable<PDFObject>
 		putDictionaryStringPDFObject(name, obj);
 	}
 
+	public void put(String name, Rect r) {
+		putDictionaryStringRect(name, r);
+	}
+
+	public void put(String name, Matrix m) {
+		putDictionaryStringMatrix(name, m);
+	}
+
+	public void put(String name, Date time) {
+		putDictionaryStringDate(name, time.getTime());
+	}
+
 	public void put(PDFObject name, boolean b) {
 		putDictionaryPDFObjectBoolean(name, b);
 	}
@@ -301,4 +316,6 @@ public class PDFObject implements Iterable<PDFObject>
 			throw new UnsupportedOperationException();
 		}
 	}
+
+	public native boolean isFilespec();
 }
