@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 package com.artifex.mupdf.fitz.android;
 
@@ -37,19 +37,31 @@ public final class AndroidDrawDevice extends NativeDevice
 		Context.init();
 	}
 
-	private native long newNative(Bitmap bitmap, int xOrigin, int yOrigin, int patchX0, int patchY0, int patchX1, int patchY1);
+	private native long newNative(Bitmap bitmap, int xOrigin, int yOrigin, int patchX0, int patchY0, int patchX1, int patchY1, boolean clear);
+
+	public AndroidDrawDevice(Bitmap bitmap, int xOrigin, int yOrigin, int patchX0, int patchY0, int patchX1, int patchY1, boolean clear) {
+		super(0);
+		pointer = newNative(bitmap, xOrigin, yOrigin, patchX0, patchY0, patchX1, patchY1, clear);
+	}
 
 	public AndroidDrawDevice(Bitmap bitmap, int xOrigin, int yOrigin, int patchX0, int patchY0, int patchX1, int patchY1) {
-		super(0);
-		pointer = newNative(bitmap, xOrigin, yOrigin, patchX0, patchY0, patchX1, patchY1);
+		this(bitmap, xOrigin, yOrigin, patchX0, patchY0, patchX1, patchY1, true);
+	}
+
+	public AndroidDrawDevice(Bitmap bitmap, int xOrigin, int yOrigin, boolean clear) {
+		this(bitmap, xOrigin, yOrigin, 0, 0, bitmap.getWidth(), bitmap.getHeight(), clear);
 	}
 
 	public AndroidDrawDevice(Bitmap bitmap, int xOrigin, int yOrigin) {
-		this(bitmap, xOrigin, yOrigin, 0, 0, bitmap.getWidth(), bitmap.getHeight());
+		this(bitmap, xOrigin, yOrigin, 0, 0, bitmap.getWidth(), bitmap.getHeight(), true);
+	}
+
+	public AndroidDrawDevice(Bitmap bitmap, boolean clear) {
+		this(bitmap, 0, 0, clear);
 	}
 
 	public AndroidDrawDevice(Bitmap bitmap) {
-		this(bitmap, 0, 0);
+		this(bitmap, 0, 0, true);
 	}
 
 	public static Bitmap drawPage(Page page, Matrix ctm) {
@@ -58,9 +70,12 @@ public final class AndroidDrawDevice extends NativeDevice
 		int h = ibox.y1 - ibox.y0;
 		Bitmap bmp = Bitmap.createBitmap(w, h, Bitmap.Config.ARGB_8888);
 		AndroidDrawDevice dev = new AndroidDrawDevice(bmp, ibox.x0, ibox.y0);
-		page.run(dev, ctm, null);
-		dev.close();
-		dev.destroy();
+		try {
+			page.run(dev, ctm, null);
+			dev.close();
+		} finally {
+			dev.destroy();
+		}
 		return bmp;
 	}
 
