@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2023 Artifex Software, Inc.
+// Copyright (C) 2004-2024 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -92,7 +92,10 @@ fz_new_buffer_from_shared_data(fz_context *ctx, const unsigned char *data, size_
 fz_buffer *
 fz_new_buffer_from_copied_data(fz_context *ctx, const unsigned char *data, size_t size)
 {
-	fz_buffer *b = fz_new_buffer(ctx, size);
+	fz_buffer *b;
+	if (size > 0 && data == NULL)
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "no data provided");
+	b = fz_new_buffer(ctx, size);
 	b->len = size;
 	memcpy(b->data, data, size);
 	return b;
@@ -211,7 +214,7 @@ void
 fz_resize_buffer(fz_context *ctx, fz_buffer *buf, size_t size)
 {
 	if (buf->shared)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot resize a buffer with shared storage");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "cannot resize a buffer with shared storage");
 	buf->data = fz_realloc(ctx, buf->data, size);
 	buf->cap = size;
 	if (buf->len > buf->cap)
