@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2023 Artifex Software, Inc.
+// Copyright (C) 2004-2024 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -23,6 +23,12 @@
 #include "mupdf/fitz.h"
 
 #include <string.h>
+#ifndef _WIN32
+#include <unistd.h> /* For unlink */
+#endif
+#include <errno.h>
+
+static void fz_reap_dead_pages(fz_context *ctx, fz_document *doc);
 
 enum
 {
@@ -33,6 +39,115 @@ enum
 #define DEFH (600)
 #define DEFEM (12)
 
+static fz_output *
+fz_new_output_to_tempfile(fz_context *ctx, char **namep)
+{
+	fz_output *out = NULL;
+#ifdef _WIN32
+	char namebuf[L_tmpnam];
+	int attempts = 0;
+#else
+	char namebuf[] = "/tmp/fztmpXXXXXX";
+#endif
+
+
+	fz_var(out);
+
+#ifdef _WIN32
+	/* Windows has no mkstemp command, so we have to use the old-style
+	 * tmpnam based system, and retry in the case of races. */
+	do
+	{
+		if (tmpnam(namebuf) == NULL)
+			fz_throw(ctx, FZ_ERROR_SYSTEM, "tmpnam failed");
+		fz_try(ctx)
+			out = fz_new_output_with_path(ctx, namebuf, 0);
+		fz_catch(ctx)
+		{
+			/* We might hit a race condition and not be able to
+			 * open the file because someone beats us to it. We'd
+			 * be unbearably unlucky to hit this 10 times in a row. */
+			attempts++;
+			if (attempts >= 10)
+				fz_rethrow(ctx);
+			else
+				fz_ignore_error(ctx);
+		}
+	}
+	while (out == NULL);
+#else
+	{
+		FILE *file;
+		int fd = mkstemp(namebuf);
+
+		if (fd == -1)
+			fz_throw(ctx, FZ_ERROR_SYSTEM, "Cannot mkstemp: %s", strerror(errno));
+		file = fdopen(fd, "w");
+		if (file == NULL)
+			fz_throw(ctx, FZ_ERROR_SYSTEM, "Failed to open temporary file");
+		out = fz_new_output_with_file_ptr(ctx, file);
+	}
+#endif
+
+	if (namep)
+	{
+		fz_try(ctx)
+			*namep = fz_strdup(ctx, namebuf);
+		fz_catch(ctx)
+		{
+			fz_drop_output(ctx, out);
+			unlink(namebuf);
+			fz_rethrow(ctx);
+		}
+	}
+
+	return out;
+}
+
+static char *
+fz_new_tmpfile_from_stream(fz_context *ctx, fz_stream *stm)
+{
+	char *name;
+	fz_output *out = fz_new_output_to_tempfile(ctx, &name);
+
+	fz_try(ctx)
+	{
+		fz_write_stream(ctx, out, stm);
+		fz_close_output(ctx, out);
+	}
+	fz_always(ctx)
+		fz_drop_output(ctx, out);
+	fz_catch(ctx)
+	{
+		fz_free(ctx, name);
+		fz_rethrow(ctx);
+	}
+
+	return name;
+}
+
+static fz_stream *
+fz_file_backed_stream(fz_context *ctx, fz_stream *stream)
+{
+	const char *oname = fz_stream_filename(ctx, stream);
+	char *name;
+
+	/* If the file has a name, it's already a file-backed stream.*/
+	if (oname)
+		return stream;
+
+	/* Otherwise we need to make it one. */
+	name = fz_new_tmpfile_from_stream(ctx, stream);
+	fz_try(ctx)
+		stream = fz_open_file_autodelete(ctx, name);
+	fz_always(ctx)
+		fz_free(ctx, name);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+
+	return stream;
+}
+
 struct fz_document_handler_context
 {
 	int refs;
@@ -55,9 +170,22 @@ fz_document_handler_context *fz_keep_document_handler_context(fz_context *ctx)
 
 void fz_drop_document_handler_context(fz_context *ctx)
 {
-	if (!ctx)
+	int i;
+
+	if (!ctx || !ctx->handler)
 		return;
 
+	for (i = 0; i < ctx->handler->count; i++)
+	{
+		if (ctx->handler->handler[i]->fin)
+		{
+			fz_try(ctx)
+				ctx->handler->handler[i]->fin(ctx, ctx->handler->handler[i]);
+			fz_catch(ctx)
+				fz_ignore_error(ctx);
+		}
+	}
+
 	if (fz_drop_imp(ctx, ctx->handler, &ctx->handler->refs))
 	{
 		fz_free(ctx, ctx->handler);
@@ -75,28 +203,43 @@ void fz_register_document_handler(fz_context *ctx, const fz_document_handler *ha
 
 	dc = ctx->handler;
 	if (dc == NULL)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Document handler list not found");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Document handler list not found");
 
 	for (i = 0; i < dc->count; i++)
 		if (dc->handler[i] == handler)
 			return;
 
 	if (dc->count >= FZ_DOCUMENT_HANDLER_MAX)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Too many document handlers");
+		fz_throw(ctx, FZ_ERROR_LIMIT, "Too many document handlers");
 
 	dc->handler[dc->count++] = handler;
 }
 
 const fz_document_handler *
 fz_recognize_document_stream_content(fz_context *ctx, fz_stream *stream, const char *magic)
+{
+	return fz_recognize_document_stream_and_dir_content(ctx, stream, NULL, magic);
+}
+
+const fz_document_handler *
+do_recognize_document_stream_and_dir_content(fz_context *ctx, fz_stream **streamp, fz_archive *dir, const char *magic, void **handler_state, fz_document_recognize_state_free_fn **handler_free_state)
 {
 	fz_document_handler_context *dc;
 	int i, best_score, best_i;
+	void *best_state = NULL;
+	fz_document_recognize_state_free_fn *best_free_state = NULL;
 	const char *ext;
+	int drop_stream = 0;
+	fz_stream *stream = *streamp;
+
+	if (handler_state)
+		*handler_state = NULL;
+	if (handler_free_state)
+		*handler_free_state = NULL;
 
 	dc = ctx->handler;
 	if (dc->count == 0)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "No document handlers registered");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "No document handlers registered");
 
 	ext = strrchr(magic, '.');
 	if (ext)
@@ -107,85 +250,187 @@ fz_recognize_document_stream_content(fz_context *ctx, fz_stream *stream, const c
 	best_score = 0;
 	best_i = -1;
 
-	if (stream && stream->seek != NULL)
+	/* If we're handed a stream, check to see if any of our document handlers
+	 * need a file. If so, change the stream to be a file-backed one. */
+	if (stream)
 	{
+		int wants_file = 0;
 		for (i = 0; i < dc->count; i++)
+			wants_file |= dc->handler[i]->wants_file;
+
+		/* Convert the stream into a file_backed stream. */
+		if (wants_file)
 		{
-			int score = 0;
-
-			fz_seek(ctx, stream, 0, SEEK_SET);
-
-			if (dc->handler[i]->recognize_content)
-				score = dc->handler[i]->recognize_content(ctx, stream);
-
-			if (best_score < score)
+			fz_stream *stream2 = fz_file_backed_stream(ctx, stream);
+			if (stream2 != stream)
 			{
-				best_score = score;
-				best_i = i;
+				/* Either we need to pass this back to our caller, or we
+				 * need to drop it. */
+				drop_stream = 1;
+				stream = stream2;
 			}
 		}
 	}
 
-	if (best_score < 100)
+	fz_try(ctx)
 	{
-		for (i = 0; i < dc->count; i++)
+		if ((stream && stream->seek != NULL) || (stream == NULL && dir != NULL))
 		{
-			int score = 0;
-			const char **entry;
-
-			if (dc->handler[i]->recognize)
-				score = dc->handler[i]->recognize(ctx, magic);
-
-			for (entry = &dc->handler[i]->mimetypes[0]; *entry; entry++)
-				if (!fz_strcasecmp(magic, *entry) && score < 100)
-				{
-					score = 100;
-					break;
-				}
-
-			if (ext)
+			for (i = 0; i < dc->count; i++)
 			{
-				for (entry = &dc->handler[i]->extensions[0]; *entry; entry++)
-					if (!fz_strcasecmp(ext, *entry) && score < 100)
+				void *state = NULL;
+				fz_document_recognize_state_free_fn *free_state = NULL;
+				int score = 0;
+
+				if (dc->handler[i]->recognize_content)
+				{
+					if (stream)
+						fz_seek(ctx, stream, 0, SEEK_SET);
+					fz_try(ctx)
+					{
+						score = dc->handler[i]->recognize_content(ctx, dc->handler[i], stream, dir, &state, &free_state);
+					}
+					fz_catch(ctx)
+					{
+						/* in case of zip errors when recognizing EPUB/XPS/DOCX files */
+						fz_rethrow_unless(ctx, FZ_ERROR_FORMAT);
+						(void)fz_convert_error(ctx, NULL); /* ugly hack to silence the error message */
+						score = 0;
+					}
+				}
+				if (best_score < score)
+				{
+					best_score = score;
+					best_i = i;
+					if (best_free_state)
+						best_free_state(ctx, best_state);
+					best_free_state = free_state;
+					best_state = state;
+				}
+				else if (free_state)
+					free_state(ctx, state);
+			}
+			if (stream)
+				fz_seek(ctx, stream, 0, SEEK_SET);
+		}
+
+		if (best_score < 100)
+		{
+			for (i = 0; i < dc->count; i++)
+			{
+				int score = 0;
+				const char **entry;
+
+				if (dc->handler[i]->recognize)
+					score = dc->handler[i]->recognize(ctx, dc->handler[i], magic);
+
+				for (entry = &dc->handler[i]->mimetypes[0]; *entry; entry++)
+					if (!fz_strcasecmp(magic, *entry) && score < 100)
 					{
 						score = 100;
 						break;
 					}
-			}
 
-			if (best_score < score)
-			{
-				best_score = score;
-				best_i = i;
+				if (ext)
+				{
+					for (entry = &dc->handler[i]->extensions[0]; *entry; entry++)
+						if (!fz_strcasecmp(ext, *entry) && score < 100)
+						{
+							score = 100;
+							break;
+						}
+				}
+
+				if (best_score < score)
+				{
+					best_score = score;
+					best_i = i;
+				}
 			}
 		}
 	}
+	fz_catch(ctx)
+	{
+		if (best_free_state)
+			best_free_state(ctx, best_state);
+		if (drop_stream)
+			fz_drop_stream(ctx, stream);
+		fz_rethrow(ctx);
+	}
 
 	if (best_i < 0)
+	{
+		if (drop_stream)
+			fz_drop_stream(ctx, stream);
 		return NULL;
+	}
+
+	/* Only if we found a handler, do we make our modified stream available to the
+	 * caller. */
+	*streamp = stream;
+
+	if (handler_state && handler_free_state)
+	{
+		*handler_state = best_state;
+		*handler_free_state = best_free_state;
+	}
+	else if (best_free_state)
+		best_free_state(ctx, best_state);
 
 	return dc->handler[best_i];
 }
 
-const fz_document_handler *fz_recognize_document_content(fz_context *ctx, const char *filename)
+const fz_document_handler *
+fz_recognize_document_stream_and_dir_content(fz_context *ctx, fz_stream *stream, fz_archive *dir, const char *magic)
 {
-	fz_stream *stream  = fz_open_file(ctx, filename);
-	const fz_document_handler *handler = NULL;
+	fz_stream *stm = stream;
+	const fz_document_handler *res;
 
+	res = do_recognize_document_stream_and_dir_content(ctx, &stm, dir, magic, NULL, NULL);
+
+	if (stm != stream)
+		fz_drop_stream(ctx, stm);
+
+	return res;
+}
+
+static const fz_document_handler *do_recognize_document_content(fz_context *ctx, const char *filename, void **handler_state, fz_document_recognize_state_free_fn **handler_free_state)
+{
+	fz_stream *stream = NULL;
+	const fz_document_handler *handler = NULL;
+	fz_archive *zip = NULL;
+	fz_stream *stm;
+
+	if (fz_is_directory(ctx, filename))
+		zip = fz_open_directory(ctx, filename);
+	else
+		stream  = fz_open_file(ctx, filename);
+
+	stm = stream;
 	fz_try(ctx)
-		handler = fz_recognize_document_stream_content(ctx, stream, filename);
+		handler = do_recognize_document_stream_and_dir_content(ctx, &stm, zip, filename, handler_state, handler_free_state);
 	fz_always(ctx)
+	{
+		if (stm != stream)
+			fz_drop_stream(ctx, stm);
 		fz_drop_stream(ctx, stream);
+		fz_drop_archive(ctx, zip);
+	}
 	fz_catch(ctx)
 		fz_rethrow(ctx);
 
 	return handler;
 }
 
+const fz_document_handler *fz_recognize_document_content(fz_context* ctx, const char* filename)
+{
+	return do_recognize_document_content(ctx, filename, NULL, NULL);
+}
+
 const fz_document_handler *
 fz_recognize_document(fz_context *ctx, const char *magic)
 {
-	return fz_recognize_document_stream_content(ctx, NULL, magic);
+	return fz_recognize_document_stream_and_dir_content(ctx, NULL, NULL, magic);
 }
 
 #if FZ_ENABLE_PDF
@@ -193,33 +438,43 @@ extern fz_document_handler pdf_document_handler;
 #endif
 
 fz_document *
-fz_open_accelerated_document_with_stream(fz_context *ctx, const char *magic, fz_stream *stream, fz_stream *accel)
+fz_open_accelerated_document_with_stream_and_dir(fz_context *ctx, const char *magic, fz_stream *stream, fz_stream *accel, fz_archive *dir)
 {
 	const fz_document_handler *handler;
+	fz_stream *wrapped_stream = stream;
+	fz_document *ret;
+	void *state = NULL;
+	fz_document_recognize_state_free_fn *free_state = NULL;
 
-	if (stream == NULL)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "no document to open");
+	if (stream == NULL && dir == NULL)
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "no document to open");
 	if (magic == NULL)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "missing file type");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "missing file type");
 
-	handler = fz_recognize_document_stream_content(ctx, stream, magic);
+	/* If this finds a handler, then this might wrap stream. If it does, we reuse the wrapped one in
+	 * the open call (hence avoiding us having to 'file-back' a stream twice), but we must free it. */
+	handler = do_recognize_document_stream_and_dir_content(ctx, &wrapped_stream, dir, magic, &state, &free_state);
 	if (!handler)
-#if FZ_ENABLE_PDF
-		handler = &pdf_document_handler;
-#else
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot find document handler for file type: '%s'", magic);
-#endif
-	if (handler->open_accel_with_stream)
-		if (accel || handler->open_with_stream == NULL)
-			return handler->open_accel_with_stream(ctx, stream, accel);
-	if (accel)
+		fz_throw(ctx, FZ_ERROR_UNSUPPORTED, "cannot find document handler for file type: '%s'", magic);
+	fz_try(ctx)
+		ret = handler->open(ctx, handler, wrapped_stream, accel, dir, state);
+	fz_always(ctx)
 	{
-		/* We've had an accelerator passed to a format that doesn't
-		 * handle it. This should never happen, as how did the
-		 * accelerator get created? */
-		fz_drop_stream(ctx, accel);
+		if (wrapped_stream != stream)
+			fz_drop_stream(ctx, wrapped_stream);
+		if (free_state && state)
+			free_state(ctx, state);
 	}
-	return handler->open_with_stream(ctx, stream);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+
+	return ret;
+}
+
+fz_document *
+fz_open_accelerated_document_with_stream(fz_context *ctx, const char *magic, fz_stream *stream, fz_stream *accel)
+{
+	return fz_open_accelerated_document_with_stream_and_dir(ctx, magic, stream, accel, NULL);
 }
 
 fz_document *
@@ -228,6 +483,12 @@ fz_open_document_with_stream(fz_context *ctx, const char *magic, fz_stream *stre
 	return fz_open_accelerated_document_with_stream(ctx, magic, stream, NULL);
 }
 
+fz_document *
+fz_open_document_with_stream_and_dir(fz_context *ctx, const char *magic, fz_stream *stream, fz_archive *dir)
+{
+	return fz_open_accelerated_document_with_stream_and_dir(ctx, magic, stream, NULL, dir);
+}
+
 fz_document *
 fz_open_document_with_buffer(fz_context *ctx, const char *magic, fz_buffer *buffer)
 {
@@ -246,51 +507,57 @@ fz_document *
 fz_open_accelerated_document(fz_context *ctx, const char *filename, const char *accel)
 {
 	const fz_document_handler *handler;
-	fz_stream *file;
+	fz_stream *file = NULL;
 	fz_stream *afile = NULL;
 	fz_document *doc = NULL;
-
-	fz_var(afile);
+	fz_archive *dir = NULL;
+	char dirname[PATH_MAX];
+	void *state = NULL;
+	fz_document_recognize_state_free_fn *free_state = NULL;
 
 	if (filename == NULL)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "no document to open");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "no document to open");
 
-	handler = fz_recognize_document_content(ctx, filename);
-	if (!handler)
-#if FZ_ENABLE_PDF
-		handler = &pdf_document_handler;
-#else
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot find document handler for file: %s", filename);
-#endif
+	if (fz_is_directory(ctx, filename))
+	{
+		/* Cannot accelerate directories, currently. */
+		dir = fz_open_directory(ctx, filename);
 
-	if (accel) {
-		if (handler->open_accel)
-			return handler->open_accel(ctx, filename, accel);
-		if (handler->open_accel_with_stream == NULL)
-		{
-			/* We're not going to be able to use the accelerator - this
-			 * should never happen, as how can one have been created? */
-			accel = NULL;
-		}
+		fz_try(ctx)
+			doc = fz_open_accelerated_document_with_stream_and_dir(ctx, filename, NULL, NULL, dir);
+		fz_always(ctx)
+			fz_drop_archive(ctx, dir);
+		fz_catch(ctx)
+			fz_rethrow(ctx);
+
+		return doc;
 	}
-	if (!accel && handler->open)
-		return handler->open(ctx, filename);
 
-	file = fz_open_file(ctx, filename);
+	handler = do_recognize_document_content(ctx, filename, &state, &free_state);
+	if (!handler)
+		fz_throw(ctx, FZ_ERROR_UNSUPPORTED, "cannot find document handler for file: %s", filename);
+
+	fz_var(afile);
+	fz_var(file);
 
 	fz_try(ctx)
 	{
-		if (accel || handler->open_with_stream == NULL)
+		file = fz_open_file(ctx, filename);
+
+		if (accel)
+			afile = fz_open_file(ctx, accel);
+		if (handler->wants_dir)
 		{
-			if (accel)
-				afile = fz_open_file(ctx, accel);
-			doc = handler->open_accel_with_stream(ctx, file, afile);
+			fz_dirname(dirname, filename, sizeof dirname);
+			dir = fz_open_directory(ctx, dirname);
 		}
-		else
-			doc = handler->open_with_stream(ctx, file);
+		doc = handler->open(ctx, handler, file, afile, dir, state);
 	}
 	fz_always(ctx)
 	{
+		if (free_state)
+			free_state(ctx, state);
+		fz_drop_archive(ctx, dir);
 		fz_drop_stream(ctx, afile);
 		fz_drop_stream(ctx, file);
 	}
@@ -323,7 +590,7 @@ void fz_output_accelerator(fz_context *ctx, fz_document *doc, fz_output *accel)
 	if (doc->output_accelerator == NULL)
 	{
 		fz_drop_output(ctx, accel);
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Document does not support writing an accelerator");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Document does not support writing an accelerator");
 	}
 
 	doc->output_accelerator(ctx, doc, accel);
@@ -355,6 +622,7 @@ fz_drop_document(fz_context *ctx, fz_document *doc)
 {
 	if (fz_drop_imp(ctx, doc, &doc->refs))
 	{
+		fz_reap_dead_pages(ctx, doc);
 		if (doc->open)
 			fz_warn(ctx, "There are still open pages in the document!");
 		if (doc->drop_document)
@@ -456,7 +724,7 @@ fz_format_link_uri(fz_context *ctx, fz_document *doc, fz_link_dest dest)
 {
 	if (doc && doc->format_link_uri)
 		return doc->format_link_uri(ctx, doc, dest);
-	fz_throw(ctx, FZ_ERROR_GENERIC, "cannot create internal links for this document type");
+	fz_throw(ctx, FZ_ERROR_ARGUMENT, "cannot create internal links for this document type");
 }
 
 fz_location
@@ -518,7 +786,7 @@ fz_load_page(fz_context *ctx, fz_document *doc, int number)
 			return fz_load_chapter_page(ctx, doc, i, number - start);
 		start += m;
 	}
-	fz_throw(ctx, FZ_ERROR_GENERIC, "Page not found: %d", number+1);
+	fz_throw(ctx, FZ_ERROR_ARGUMENT, "invalid page number: %d", number+1);
 }
 
 fz_location fz_last_page(fz_context *ctx, fz_document *doc)
@@ -629,6 +897,28 @@ fz_document_output_intent(fz_context *ctx, fz_document *doc)
 	return NULL;
 }
 
+static void
+fz_reap_dead_pages(fz_context *ctx, fz_document *doc)
+{
+	fz_page *page;
+	fz_page *next_page;
+
+	for (page = doc->open; page; page = next_page)
+	{
+		next_page = page->next;
+		if (!page->doc)
+		{
+			if (page->next != NULL)
+				page->next->prev = page->prev;
+			if (page->prev != NULL)
+				*page->prev = page->next;
+			fz_free(ctx, page);
+			if (page == doc->open)
+				doc->open = next_page;
+		}
+	}
+}
+
 fz_page *
 fz_load_chapter_page(fz_context *ctx, fz_document *doc, int chapter, int number)
 {
@@ -639,17 +929,19 @@ fz_load_chapter_page(fz_context *ctx, fz_document *doc, int chapter, int number)
 
 	fz_ensure_layout(ctx, doc);
 
+	// Trigger reaping dead pages when we load a new page.
+	fz_reap_dead_pages(ctx, doc);
+
 	/* Protect modifications to the page list to cope with
 	 * destruction of pages on other threads. */
-	fz_lock(ctx, FZ_LOCK_ALLOC);
 	for (page = doc->open; page; page = page->next)
+	{
 		if (page->chapter == chapter && page->number == number)
 		{
-			fz_keep_page_locked(ctx, page);
-			fz_unlock(ctx, FZ_LOCK_ALLOC);
+			fz_keep_page(ctx, page);
 			return page;
 		}
-	fz_unlock(ctx, FZ_LOCK_ALLOC);
+	}
 
 	if (doc->load_page)
 	{
@@ -660,12 +952,10 @@ fz_load_chapter_page(fz_context *ctx, fz_document *doc, int chapter, int number)
 		/* Insert new page at the head of the list of open pages. */
 		if (!page->incomplete)
 		{
-			fz_lock(ctx, FZ_LOCK_ALLOC);
 			if ((page->next = doc->open) != NULL)
 				doc->open->prev = &page->next;
 			doc->open = page;
 			page->prev = &doc->open;
-			fz_unlock(ctx, FZ_LOCK_ALLOC);
 		}
 		return page;
 	}
@@ -697,6 +987,24 @@ fz_bound_page_box(fz_context *ctx, fz_page *page, fz_box_type box)
 	return fz_empty_rect;
 }
 
+void
+fz_run_document_structure(fz_context *ctx, fz_document *doc, fz_device *dev, fz_cookie *cookie)
+{
+	if (doc && doc->run_structure)
+	{
+		fz_try(ctx)
+		{
+			doc->run_structure(ctx, doc, dev, cookie);
+		}
+		fz_catch(ctx)
+		{
+			dev->close_device = NULL; /* aborted run, don't warn about unclosed device */
+			fz_rethrow_unless(ctx, FZ_ERROR_ABORT);
+			fz_ignore_error(ctx);
+		}
+	}
+}
+
 void
 fz_run_page_contents(fz_context *ctx, fz_page *page, fz_device *dev, fz_matrix transform, fz_cookie *cookie)
 {
@@ -709,8 +1017,8 @@ fz_run_page_contents(fz_context *ctx, fz_page *page, fz_device *dev, fz_matrix t
 		fz_catch(ctx)
 		{
 			dev->close_device = NULL; /* aborted run, don't warn about unclosed device */
-			if (fz_caught(ctx) != FZ_ERROR_ABORT)
-				fz_rethrow(ctx);
+			fz_rethrow_unless(ctx, FZ_ERROR_ABORT);
+			fz_ignore_error(ctx);
 		}
 	}
 }
@@ -727,8 +1035,8 @@ fz_run_page_annots(fz_context *ctx, fz_page *page, fz_device *dev, fz_matrix tra
 		fz_catch(ctx)
 		{
 			dev->close_device = NULL; /* aborted run, don't warn about unclosed device */
-			if (fz_caught(ctx) != FZ_ERROR_ABORT)
-				fz_rethrow(ctx);
+			fz_rethrow_unless(ctx, FZ_ERROR_ABORT);
+			fz_ignore_error(ctx);
 		}
 	}
 }
@@ -745,8 +1053,8 @@ fz_run_page_widgets(fz_context *ctx, fz_page *page, fz_device *dev, fz_matrix tr
 		fz_catch(ctx)
 		{
 			dev->close_device = NULL; /* aborted run, don't warn about unclosed device */
-			if (fz_caught(ctx) != FZ_ERROR_ABORT)
-				fz_rethrow(ctx);
+			fz_rethrow_unless(ctx, FZ_ERROR_ABORT);
+			fz_ignore_error(ctx);
 		}
 	}
 }
@@ -774,31 +1082,22 @@ fz_keep_page(fz_context *ctx, fz_page *page)
 	return fz_keep_imp(ctx, page, &page->refs);
 }
 
-fz_page *
-fz_keep_page_locked(fz_context *ctx, fz_page *page)
-{
-	return fz_keep_imp_locked(ctx, page, &page->refs);
-}
-
 void
 fz_drop_page(fz_context *ctx, fz_page *page)
 {
 	if (fz_drop_imp(ctx, page, &page->refs))
 	{
-		/* Remove page from the list of open pages */
-		fz_lock(ctx, FZ_LOCK_ALLOC);
-		if (page->next != NULL)
-			page->next->prev = page->prev;
-		if (page->prev != NULL)
-			*page->prev = page->next;
-		fz_unlock(ctx, FZ_LOCK_ALLOC);
+		fz_document *doc = page->doc;
 
 		if (page->drop_page)
 			page->drop_page(ctx, page);
 
-		fz_drop_document(ctx, page->doc);
+		// Mark the page as dead so we can reap the struct allocation later.
+		page->doc = NULL;
+		page->chapter = -1;
+		page->number = -1;
 
-		fz_free(ctx, page);
+		fz_drop_document(ctx, doc);
 	}
 }
 
@@ -835,7 +1134,7 @@ fz_link *fz_create_link(fz_context *ctx, fz_page *page, fz_rect bbox, const char
 	if (page == NULL || uri == NULL)
 		return NULL;
 	if (page->create_link == NULL)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "This format of document does not support creating links");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "This format of document does not support creating links");
 	return page->create_link(ctx, page, bbox, uri);
 }
 
@@ -844,7 +1143,7 @@ void fz_delete_link(fz_context *ctx, fz_page *page, fz_link *link)
 	if (page == NULL || link == NULL)
 		return;
 	if (page->delete_link == NULL)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "This format of document does not support deleting links");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "This format of document does not support deleting links");
 	page->delete_link(ctx, page, link);
 }
 
@@ -853,7 +1152,7 @@ void fz_set_link_rect(fz_context *ctx, fz_link *link, fz_rect rect)
 	if (link == NULL)
 		return;
 	if (link->set_rect_fn == NULL)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "This format of document does not support updating link bounds");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "This format of document does not support updating link bounds");
 	link->set_rect_fn(ctx, link, rect);
 }
 
@@ -862,7 +1161,7 @@ void fz_set_link_uri(fz_context *ctx, fz_link *link, const char *uri)
 	if (link == NULL)
 		return;
 	if (link->set_uri_fn == NULL)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "This format of document does not support updating link uri");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "This format of document does not support updating link uri");
 	link->set_uri_fn(ctx, link, uri);
 }
 
@@ -870,51 +1169,20 @@ void *
 fz_process_opened_pages(fz_context *ctx, fz_document *doc, fz_process_opened_page_fn *process_opened_page, void *state)
 {
 	fz_page *page;
-	fz_page *kept = NULL;
-	fz_page *dropme = NULL;
-	void *ret = NULL;
+	void *ret;
 
-	fz_var(kept);
-	fz_var(dropme);
-	fz_var(page);
-	fz_try(ctx)
+	for (page = doc->open; page != NULL; page = page->next)
 	{
-		/* We can only walk the page list while the alloc lock is taken, so gymnastics are required. */
-		/* Loop invariant: at any point where we might throw, kept != NULL iff we are unlocked. */
-		fz_lock(ctx, FZ_LOCK_ALLOC);
-		for (page = doc->open; ret == NULL && page != NULL; page = page->next)
-		{
-			/* Keep an extra reference to the page so that no other thread can remove it. */
-			kept = fz_keep_page_locked(ctx, page);
-			fz_unlock(ctx, FZ_LOCK_ALLOC);
-			/* Drop any extra reference we might still have to a previous page. */
-			fz_drop_page(ctx, dropme);
-			dropme = NULL;
+		// Skip dead pages.
+		if (page->doc == NULL)
+			continue;
 
-			ret = process_opened_page(ctx, page, state);
-
-			/* We can't drop kept here, because that would give us a race condition with
-			 * us taking the lock and hoping that 'page' would still be valid. So remember it
-			 * for dropping later. */
-			dropme = kept;
-			kept = NULL;
-			fz_lock(ctx, FZ_LOCK_ALLOC);
-		}
-		/* unlock (and final drop of dropme) happens in the always. */
-	}
-	fz_always(ctx)
-	{
-		if (kept == NULL)
-			fz_unlock(ctx, FZ_LOCK_ALLOC);
-		fz_drop_page(ctx, kept);
-		fz_drop_page(ctx, dropme);
-	}
-	fz_catch(ctx)
-	{
-		fz_rethrow(ctx);
+		ret = process_opened_page(ctx, page, state);
+		if (ret)
+			return ret;
 	}
 
-	return ret;
+	return NULL;
 }
 
 const char *
