@@ -387,6 +387,9 @@ const fz_irect fz_empty_irect = { FZ_MAX_INF_RECT, FZ_MAX_INF_RECT, FZ_MIN_INF_R
 const fz_irect fz_invalid_irect = { 0, 0, -1, -1 };
 const fz_irect fz_unit_bbox = { 0, 0, 1, 1 };
 
+const fz_quad fz_infinite_quad = { { -INFINITY, INFINITY}, {INFINITY, INFINITY}, {-INFINITY, -INFINITY}, {INFINITY, -INFINITY} };
+const fz_quad fz_invalid_quad = { {NAN, NAN}, {NAN, NAN}, {NAN, NAN}, {NAN, NAN} };
+
 fz_irect
 fz_irect_from_rect(fz_rect r)
 {
@@ -640,10 +643,86 @@ int fz_contains_rect(fz_rect a, fz_rect b)
 		(a.y1 >= b.y1));
 }
 
+int
+fz_is_valid_quad(fz_quad q)
+{
+	return (!isnan(q.ll.x) &&
+		!isnan(q.ll.y) &&
+		!isnan(q.ul.x) &&
+		!isnan(q.ul.y) &&
+		!isnan(q.lr.x) &&
+		!isnan(q.lr.y) &&
+		!isnan(q.ur.x) &&
+		!isnan(q.ur.y));
+}
+
+int
+fz_is_infinite_quad(fz_quad q)
+{
+	/* For a quad to be infinite, all the ordinates need to be infinite. */
+	if (!isinf(q.ll.x) ||
+		!isinf(q.ll.y) ||
+		!isinf(q.ul.x) ||
+		!isinf(q.ul.y) ||
+		!isinf(q.lr.x) ||
+		!isinf(q.lr.y) ||
+		!isinf(q.ur.x) ||
+		!isinf(q.ur.y))
+		return 0;
+
+	/* Just because all the ordinates are infinite, we don't necessarily have an infinite quad. */
+	/* The quad points in order are: ll, ul, ur, lr  OR  reversed:  ll, lr, ur, ul */
+	/* The required points are: (-inf, -inf) (-inf, inf) (inf, inf) (inf, -inf) */
+
+	/* Not the fastest way to code it, but easy to understand! */
+#define INF_QUAD_TEST(A,B,C,D) \
+	if (q.A.x < 0 && q.A.y < 0 && q.B.x < 0 && q.B.y > 0 && q.C.x > 0 && q.C.y > 0 && q.D.x > 0 && q.D.y < 0) return 1
+
+	INF_QUAD_TEST(ll, ul, ur, lr);
+	INF_QUAD_TEST(ul, ur, lr, ll);
+	INF_QUAD_TEST(ur, lr, ll, ul);
+	INF_QUAD_TEST(lr, ll, ul, ur);
+	INF_QUAD_TEST(ll, lr, ur, ul);
+	INF_QUAD_TEST(lr, ur, ul, ll);
+	INF_QUAD_TEST(ur, ul, ll, lr);
+	INF_QUAD_TEST(ul, ll, lr, ur);
+#undef INF_QUAD_TEST
+
+	return 0;
+}
+
+int fz_is_empty_quad(fz_quad q)
+{
+	/* Using "Shoelace" formula */
+	float area;
+
+	if (fz_is_infinite_quad(q))
+		return 0;
+	if (!fz_is_valid_quad(q))
+		return 1; /* All invalid quads are empty */
+
+	area = q.ll.x * q.lr.y +
+		q.lr.x * q.ur.y +
+		q.ur.x * q.ul.y +
+		q.ul.x * q.ll.y -
+		q.lr.x * q.ll.y -
+		q.ur.x * q.lr.y -
+		q.ul.x * q.ur.y -
+		q.ll.x * q.ul.y;
+
+	return area == 0;
+}
+
 fz_rect
 fz_rect_from_quad(fz_quad q)
 {
 	fz_rect r;
+
+	if (!fz_is_valid_quad(q))
+		return fz_invalid_rect;
+	if (fz_is_infinite_quad(q))
+		return fz_infinite_rect;
+
 	r.x0 = MIN4(q.ll.x, q.lr.x, q.ul.x, q.ur.x);
 	r.y0 = MIN4(q.ll.y, q.lr.y, q.ul.y, q.ur.y);
 	r.x1 = MAX4(q.ll.x, q.lr.x, q.ul.x, q.ur.x);
@@ -654,6 +733,11 @@ fz_rect_from_quad(fz_quad q)
 fz_quad
 fz_transform_quad(fz_quad q, fz_matrix m)
 {
+	if (!fz_is_valid_quad(q))
+		return q;
+	if (fz_is_infinite_quad(q))
+		return q;
+
 	q.ul = fz_transform_point(q.ul, m);
 	q.ur = fz_transform_point(q.ur, m);
 	q.ll = fz_transform_point(q.ll, m);
@@ -665,6 +749,12 @@ fz_quad
 fz_quad_from_rect(fz_rect r)
 {
 	fz_quad q;
+
+	if (!fz_is_valid_rect(r))
+		return fz_invalid_quad;
+	if (fz_is_infinite_rect(r))
+		return fz_infinite_quad;
+
 	q.ul = fz_make_point(r.x0, r.y0);
 	q.ur = fz_make_point(r.x1, r.y0);
 	q.ll = fz_make_point(r.x0, r.y1);
@@ -672,6 +762,7 @@ fz_quad_from_rect(fz_rect r)
 	return q;
 }
 
+
 int fz_is_point_inside_rect(fz_point p, fz_rect r)
 {
 	return (p.x >= r.x0 && p.x < r.x1 && p.y >= r.y0 && p.y < r.y1);
@@ -740,6 +831,11 @@ static int fz_is_point_inside_triangle(fz_point p, fz_point a, fz_point b, fz_po
 
 int fz_is_point_inside_quad(fz_point p, fz_quad q)
 {
+	if (!fz_is_valid_quad(q))
+		return 0;
+	if (fz_is_infinite_quad(q))
+		return 1;
+
 	return
 		fz_is_point_inside_triangle(p, q.ul, q.ur, q.lr) ||
 		fz_is_point_inside_triangle(p, q.ul, q.lr, q.ll);
@@ -747,6 +843,11 @@ int fz_is_point_inside_quad(fz_point p, fz_quad q)
 
 int fz_is_quad_inside_quad(fz_quad needle, fz_quad haystack)
 {
+	if (!fz_is_valid_quad(needle) || !fz_is_valid_quad(haystack))
+		return 0;
+	if (fz_is_infinite_quad(haystack))
+		return 1;
+
 	return
 		fz_is_point_inside_quad(needle.ul, haystack) &&
 		fz_is_point_inside_quad(needle.ur, haystack) &&
