@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2023 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -17,21 +17,19 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 
-fz_link *
-fz_new_link(fz_context *ctx, fz_rect bbox, const char *uri)
-{
-	fz_link *link;
+#include <math.h>
 
-	link = fz_malloc_struct(ctx, fz_link);
+fz_link *
+fz_new_link_of_size(fz_context *ctx, int size, fz_rect rect, const char *uri)
+{
+	fz_link *link = Memento_label(fz_calloc(ctx, 1, size), "fz_link");
 	link->refs = 1;
-	link->rect = bbox;
-	link->next = NULL;
-	link->uri = NULL;
+	link->rect = rect;
 
 	fz_try(ctx)
 		link->uri = fz_strdup(ctx, uri);
@@ -56,6 +54,10 @@ fz_drop_link(fz_context *ctx, fz_link *link)
 	while (fz_drop_imp(ctx, link, &link->refs))
 	{
 		fz_link *next = link->next;
+
+		if (link->drop)
+			link->drop(ctx, link);
+
 		fz_free(ctx, link->uri);
 		fz_free(ctx, link);
 		link = next;
@@ -65,14 +67,20 @@ fz_drop_link(fz_context *ctx, fz_link *link)
 int
 fz_is_external_link(fz_context *ctx, const char *uri)
 {
+	const char *mark;
 	/* Basically, this function returns true, if the URI starts with
 	 * a valid 'scheme' followed by ':'. */
 
+	 if (!uri)
+		 return 0;
+
 	/* All schemes must start with a letter; exit if we don't. */
 	if ((*uri < 'a' || *uri > 'z') && (*uri < 'A' || *uri > 'Z'))
 		return 0;
 	uri++;
 
+	mark = uri;
+
 	/* Subsequent characters can be letters, digits, +, -, or . */
 	while ((*uri >= 'a' && *uri <= 'z') ||
 		(*uri >= 'A' && *uri <= 'Z') ||
@@ -82,17 +90,17 @@ fz_is_external_link(fz_context *ctx, const char *uri)
 		(*uri == '.'))
 		++uri;
 
-	return uri[0] == ':';
+	return uri[0] == ':' && (uri - mark) > 1;
 }
 
 fz_link_dest fz_make_link_dest_none(void)
 {
-	fz_link_dest dest = { { -1, -1 }, FZ_LINK_DEST_XYZ, 0, 0, 0, 0, 0 };
+	fz_link_dest dest = { { -1, -1 }, FZ_LINK_DEST_XYZ, NAN, NAN, NAN, NAN, NAN };
 	return dest;
 }
 
 fz_link_dest fz_make_link_dest_xyz(int chapter, int page, float x, float y, float z)
 {
-	fz_link_dest dest = { { chapter, page }, FZ_LINK_DEST_XYZ, x, y, 0, 0, z };
+	fz_link_dest dest = { { chapter, page }, FZ_LINK_DEST_XYZ, x, y, NAN, NAN, z };
 	return dest;
 }
