@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
+
 #include "mupdf/fitz.h"
 
 #include <limits.h>
@@ -230,25 +252,6 @@ const char *fz_pcl_write_options_usage =
 	"\tis_oce9050: Disable/Enable Oce 9050 model-specific output\n"
 	"\n";
 
-/*
-	Initialize PCL option struct for a given preset.
-
-	Currently defined presets include:
-
-		generic	Generic PCL printer
-		ljet4	HP DeskJet
-		dj500	HP DeskJet 500
-		fs600	Kyocera FS-600
-		lj	HP LaserJet, HP LaserJet Plus
-		lj2	HP LaserJet IIp, HP LaserJet IId
-		lj3	HP LaserJet III
-		lj3d	HP LaserJet IIId
-		lj4	HP LaserJet 4
-		lj4pl	HP LaserJet 4 PL
-		lj4d	HP LaserJet 4d
-		lp2563b	HP 2563B line printer
-		oce9050	Oce 9050 Line printer
-*/
 void fz_pcl_preset(fz_context *ctx, fz_pcl_options *opts, const char *preset)
 {
 	if (preset == NULL || *preset == 0 || !strcmp(preset, "generic"))
@@ -278,28 +281,9 @@ void fz_pcl_preset(fz_context *ctx, fz_pcl_options *opts, const char *preset)
 	else if (!strcmp(preset, "oce9050"))
 		copy_opts(opts, &fz_pcl_options_oce9050);
 	else
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Unknown preset '%s'", preset);
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Unknown preset '%s'", preset);
 }
 
-/*
-	Parse PCL options.
-
-	Currently defined options and values are as follows:
-
-		preset=X	Either "generic" or one of the presets as for fz_pcl_preset.
-		spacing=0	No vertical spacing capability
-		spacing=1	PCL 3 spacing (<ESC>*p+<n>Y)
-		spacing=2	PCL 4 spacing (<ESC>*b<n>Y)
-		spacing=3	PCL 5 spacing (<ESC>*b<n>Y and clear seed row)
-		mode2		Disable/Enable mode 2 graphics compression
-		mode3		Disable/Enable mode 3 graphics compression
-		eog_reset	End of graphics (<ESC>*rB) resets all parameters
-		has_duplex	Duplex supported (<ESC>&l<duplex>S)
-		has_papersize	Papersize setting supported (<ESC>&l<sizecode>A)
-		has_copies	Number of copies supported (<ESC>&l<copies>X)
-		is_ljet4pjl	Disable/Enable HP 4PJL model-specific output
-		is_oce9050	Disable/Enable Oce 9050 model-specific output
-*/
 fz_pcl_options *
 fz_parse_pcl_options(fz_context *ctx, fz_pcl_options *opts, const char *args)
 {
@@ -320,7 +304,7 @@ fz_parse_pcl_options(fz_context *ctx, fz_pcl_options *opts, const char *args)
 		case 1: opts->features = (opts->features & ~PCL_ANY_SPACING) | PCL3_SPACING; break;
 		case 2: opts->features = (opts->features & ~PCL_ANY_SPACING) | PCL4_SPACING; break;
 		case 3: opts->features = (opts->features & ~PCL_ANY_SPACING) | PCL5_SPACING; break;
-		default: fz_throw(ctx, FZ_ERROR_GENERIC, "Unsupported PCL spacing %d (0-3 only)", atoi(val));
+		default: fz_throw(ctx, FZ_ERROR_ARGUMENT, "Unsupported PCL spacing %d (0-3 only)", atoi(val));
 		}
 	}
 	if (fz_has_option(ctx, args, "mode2", &val))
@@ -330,7 +314,7 @@ fz_parse_pcl_options(fz_context *ctx, fz_pcl_options *opts, const char *args)
 		else if (fz_option_eq(val, "yes"))
 			opts->features |= PCL_MODE_2_COMPRESSION;
 		else
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Expected 'yes' or 'no' for mode2 value");
+			fz_throw(ctx, FZ_ERROR_ARGUMENT, "Expected 'yes' or 'no' for mode2 value");
 	}
 	if (fz_has_option(ctx, args, "mode3", &val))
 	{
@@ -339,7 +323,7 @@ fz_parse_pcl_options(fz_context *ctx, fz_pcl_options *opts, const char *args)
 		else if (fz_option_eq(val, "yes"))
 			opts->features |= PCL_MODE_3_COMPRESSION;
 		else
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Expected 'yes' or 'no' for mode3 value");
+			fz_throw(ctx, FZ_ERROR_ARGUMENT, "Expected 'yes' or 'no' for mode3 value");
 	}
 	if (fz_has_option(ctx, args, "eog_reset", &val))
 	{
@@ -348,7 +332,7 @@ fz_parse_pcl_options(fz_context *ctx, fz_pcl_options *opts, const char *args)
 		else if (fz_option_eq(val, "yes"))
 			opts->features |= PCL_END_GRAPHICS_DOES_RESET;
 		else
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Expected 'yes' or 'no' for eog_reset value");
+			fz_throw(ctx, FZ_ERROR_ARGUMENT, "Expected 'yes' or 'no' for eog_reset value");
 	}
 	if (fz_has_option(ctx, args, "has_duplex", &val))
 	{
@@ -357,7 +341,7 @@ fz_parse_pcl_options(fz_context *ctx, fz_pcl_options *opts, const char *args)
 		else if (fz_option_eq(val, "yes"))
 			opts->features |= PCL_HAS_DUPLEX;
 		else
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Expected 'yes' or 'no' for has_duplex value");
+			fz_throw(ctx, FZ_ERROR_ARGUMENT, "Expected 'yes' or 'no' for has_duplex value");
 	}
 	if (fz_has_option(ctx, args, "has_papersize", &val))
 	{
@@ -366,7 +350,7 @@ fz_parse_pcl_options(fz_context *ctx, fz_pcl_options *opts, const char *args)
 		else if (fz_option_eq(val, "yes"))
 			opts->features |= PCL_CAN_SET_PAPER_SIZE;
 		else
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Expected 'yes' or 'no' for has_papersize value");
+			fz_throw(ctx, FZ_ERROR_ARGUMENT, "Expected 'yes' or 'no' for has_papersize value");
 	}
 	if (fz_has_option(ctx, args, "has_copies", &val))
 	{
@@ -375,7 +359,7 @@ fz_parse_pcl_options(fz_context *ctx, fz_pcl_options *opts, const char *args)
 		else if (fz_option_eq(val, "yes"))
 			opts->features |= PCL_CAN_PRINT_COPIES;
 		else
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Expected 'yes' or 'no' for has_papersize value");
+			fz_throw(ctx, FZ_ERROR_ARGUMENT, "Expected 'yes' or 'no' for has_papersize value");
 	}
 	if (fz_has_option(ctx, args, "is_ljet4pjl", &val))
 	{
@@ -384,7 +368,7 @@ fz_parse_pcl_options(fz_context *ctx, fz_pcl_options *opts, const char *args)
 		else if (fz_option_eq(val, "yes"))
 			opts->features |= HACK__IS_A_LJET4PJL;
 		else
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Expected 'yes' or 'no' for is_ljet4pjl value");
+			fz_throw(ctx, FZ_ERROR_ARGUMENT, "Expected 'yes' or 'no' for is_ljet4pjl value");
 	}
 	if (fz_has_option(ctx, args, "is_oce9050", &val))
 	{
@@ -393,7 +377,7 @@ fz_parse_pcl_options(fz_context *ctx, fz_pcl_options *opts, const char *args)
 		else if (fz_option_eq(val, "yes"))
 			opts->features |= HACK__IS_A_OCE9050;
 		else
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Expected 'yes' or 'no' for is_oce9050 value");
+			fz_throw(ctx, FZ_ERROR_ARGUMENT, "Expected 'yes' or 'no' for is_oce9050 value");
 	}
 
 	return opts;
@@ -605,7 +589,7 @@ static void guess_paper_size(fz_pcl_options *pcl, int w, int h, int xres, int yr
 	h = h * 300 / xres;
 
 	/* Look for an exact match */
-	for (size = 0; size < num_elems(papersizes); size++)
+	for (size = 0; size < (int)num_elems(papersizes); size++)
 	{
 		if (papersizes[size].code > eCustomPaperSize && (pcl->features & PCL_HAS_RICOH_PAPER_SIZES) == 0)
 			continue;
@@ -632,7 +616,7 @@ static void guess_paper_size(fz_pcl_options *pcl, int w, int h, int xres, int yr
 			/* Send the next larger one (minimise waste) */
 			int i;
 			int best_waste = INT_MAX;
-			for (i = 0; i < num_elems(papersizes); i++)
+			for (i = 0; i < (int)num_elems(papersizes); i++)
 			{
 				int waste;
 				if (papersizes[i].code > eCustomPaperSize && (pcl->features & PCL_HAS_RICOH_PAPER_SIZES) == 0)
@@ -658,7 +642,7 @@ static void guess_paper_size(fz_pcl_options *pcl, int w, int h, int xres, int yr
 
 	/* Now, size = The best size we have (or num_elems(papersizes)) if it's too big */
 
-	if (size < num_elems(papersizes))
+	if (size < (int)num_elems(papersizes))
 		pcl->paper_size = papersizes[size].code;
 	else
 		pcl->paper_size = eCustomPaperSize; /* Custom */
@@ -755,6 +739,7 @@ fz_write_pixmap_as_pcl(fz_context *ctx, fz_output *out, const fz_pixmap *pixmap,
 	{
 		fz_write_header(ctx, writer, pixmap->w, pixmap->h, pixmap->n, pixmap->alpha, pixmap->xres, pixmap->yres, 0, pixmap->colorspace, pixmap->seps);
 		fz_write_band(ctx, writer, pixmap->stride, pixmap->h, pixmap->samples);
+		fz_close_band_writer(ctx, writer);
 	}
 	fz_always(ctx)
 		fz_drop_band_writer(ctx, writer);
@@ -785,13 +770,13 @@ color_pcl_write_header(fz_context *ctx, fz_band_writer *writer_, fz_colorspace *
 	int yres = writer->super.yres;
 
 	if (a != 0)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "color PCL cannot write alpha channel");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "color PCL cannot write alpha channel");
 	if (s != 0)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "color PCL cannot write spot colors");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "color PCL cannot write spot colors");
 	if (n != 3)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "color PCL must be RGB");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "color PCL must be RGB");
 
-	writer->linebuf = fz_malloc(ctx, w * 3 * 2);
+	writer->linebuf = Memento_label(fz_malloc(ctx, w * 3 * 2), "color_pcl_linebuf");
 
 	guess_paper_size(&writer->options, w, h, xres, yres);
 
@@ -1039,7 +1024,7 @@ fz_band_writer *fz_new_color_pcl_band_writer(fz_context *ctx, fz_output *out, co
  * In the worst case, the result is N+(N/127)+1 bytes long,
  * where N is the original byte count (end_row - row).
  */
-int
+static int
 mode2compress(unsigned char *out, const unsigned char *in, int in_len)
 {
 	int x;
@@ -1095,7 +1080,7 @@ mode2compress(unsigned char *out, const unsigned char *in, int in_len)
  * Returns the number of bytes stored.	In the worst case,
  * the number of bytes is bytecount+(bytecount/8)+1.
  */
-int
+static int
 mode3compress(unsigned char *out, const unsigned char *in, unsigned char *prev, int in_len)
 {
 	unsigned char *compressed = out;
@@ -1156,6 +1141,7 @@ fz_write_bitmap_as_pcl(fz_context *ctx, fz_output *out, const fz_bitmap *bitmap,
 	{
 		fz_write_header(ctx, writer, bitmap->w, bitmap->h, 1, 0, bitmap->xres, bitmap->yres, 0, NULL, NULL);
 		fz_write_band(ctx, writer, bitmap->stride, bitmap->h, bitmap->samples);
+		fz_close_band_writer(ctx, writer);
 	}
 	fz_always(ctx)
 		fz_drop_band_writer(ctx, writer);
@@ -1188,11 +1174,11 @@ mono_pcl_write_header(fz_context *ctx, fz_band_writer *writer_, fz_colorspace *c
 	int max_mode_3_size;
 
 	if (writer->super.alpha != 0)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "mono PCL cannot write alpha channel");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "mono PCL cannot write alpha channel");
 	if (writer->super.s != 0)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "mono PCL cannot write spot colors");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "mono PCL cannot write spot colors");
 	if (writer->super.n != 1)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "mono PCL must be grayscale");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "mono PCL must be grayscale");
 
 	line_size = (w + 7)/8;
 	max_mode_2_size = line_size + (line_size/127) + 1;
@@ -1459,9 +1445,7 @@ fz_save_bitmap_as_pcl(fz_context *ctx, fz_bitmap *bitmap, char *filename, int ap
 
 /* High-level document writer interface */
 
-typedef struct fz_pcl_writer_s fz_pcl_writer;
-
-struct fz_pcl_writer_s
+typedef struct
 {
 	fz_document_writer super;
 	fz_draw_options draw;
@@ -1469,7 +1453,7 @@ struct fz_pcl_writer_s
 	fz_pixmap *pixmap;
 	int mono;
 	fz_output *out;
-};
+} fz_pcl_writer;
 
 static fz_device *
 pcl_begin_page(fz_context *ctx, fz_document_writer *wri_, fz_rect mediabox)
@@ -1526,26 +1510,36 @@ pcl_drop_writer(fz_context *ctx, fz_document_writer *wri_)
 }
 
 fz_document_writer *
-fz_new_pcl_writer(fz_context *ctx, const char *path, const char *options)
+fz_new_pcl_writer_with_output(fz_context *ctx, fz_output *out, const char *options)
 {
-	fz_pcl_writer *wri = fz_new_derived_document_writer(ctx, fz_pcl_writer, pcl_begin_page, pcl_end_page, pcl_close_writer, pcl_drop_writer);
+	fz_pcl_writer *wri = NULL;
 	const char *val;
 
+	fz_var(wri);
+
 	fz_try(ctx)
 	{
+		wri = fz_new_derived_document_writer(ctx, fz_pcl_writer, pcl_begin_page, pcl_end_page, pcl_close_writer, pcl_drop_writer);
 		fz_parse_draw_options(ctx, &wri->draw, options);
 		fz_parse_pcl_options(ctx, &wri->pcl, options);
 		if (fz_has_option(ctx, options, "colorspace", &val))
 			if (fz_option_eq(val, "mono"))
 				wri->mono = 1;
-		wri->out = fz_new_output_with_path(ctx, path ? path : "out.pcl", 0);
+		wri->out = out;
 	}
 	fz_catch(ctx)
 	{
-		fz_drop_output(ctx, wri->out);
+		fz_drop_output(ctx, out);
 		fz_free(ctx, wri);
 		fz_rethrow(ctx);
 	}
 
 	return (fz_document_writer*)wri;
 }
+
+fz_document_writer *
+fz_new_pcl_writer(fz_context *ctx, const char *path, const char *options)
+{
+	fz_output *out = fz_new_output_with_path(ctx, path ? path : "out.pcl", 0);
+	return fz_new_pcl_writer_with_output(ctx, out, options);
+}
