@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2024 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -35,6 +35,7 @@
 #define ZIP_DATA_DESC_SIG 0x08074b50
 #define ZIP_CENTRAL_DIRECTORY_SIG 0x02014b50
 #define ZIP_END_OF_CENTRAL_DIRECTORY_SIG 0x06054b50
+#define ZIP_UP_SIG 0x7075
 
 #define ZIP64_END_OF_CENTRAL_DIRECTORY_LOCATOR_SIG 0x07064b50
 #define ZIP64_END_OF_CENTRAL_DIRECTORY_SIG 0x06064b50
@@ -272,7 +273,7 @@ static void read_zip_dir_imp(fz_context *ctx, fz_zip_archive *zip, int64_t start
 
 	sig = fz_read_uint32_le(ctx, file);
 	if (sig != ZIP_END_OF_CENTRAL_DIRECTORY_SIG)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "wrong zip end of central directory signature (0x%x)", sig);
+		fz_throw(ctx, FZ_ERROR_FORMAT, "wrong zip end of central directory signature (0x%x)", sig);
 
 	(void) fz_read_uint16_le(ctx, file); /* this disk */
 	(void) fz_read_uint16_le(ctx, file); /* start disk */
@@ -290,7 +291,7 @@ static void read_zip_dir_imp(fz_context *ctx, fz_zip_archive *zip, int64_t start
 
 		sig = fz_read_uint32_le(ctx, file);
 		if (sig != ZIP64_END_OF_CENTRAL_DIRECTORY_LOCATOR_SIG)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "wrong zip64 end of central directory locator signature (0x%x)", sig);
+			fz_throw(ctx, FZ_ERROR_FORMAT, "wrong zip64 end of central directory locator signature (0x%x)", sig);
 
 		(void) fz_read_uint32_le(ctx, file); /* start disk */
 		offset64 = fz_read_uint64_le(ctx, file); /* offset to end of central directory record */
@@ -299,7 +300,7 @@ static void read_zip_dir_imp(fz_context *ctx, fz_zip_archive *zip, int64_t start
 
 		sig = fz_read_uint32_le(ctx, file);
 		if (sig != ZIP64_END_OF_CENTRAL_DIRECTORY_SIG)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "wrong zip64 end of central directory signature (0x%x)", sig);
+			fz_throw(ctx, FZ_ERROR_FORMAT, "wrong zip64 end of central directory signature (0x%x)", sig);
 
 		(void) fz_read_uint64_le(ctx, file); /* size of record */
 		(void) fz_read_uint16_le(ctx, file); /* version made by */
@@ -331,7 +332,7 @@ static void read_zip_dir_imp(fz_context *ctx, fz_zip_archive *zip, int64_t start
 		{
 			sig = fz_read_uint32_le(ctx, file);
 			if (sig != ZIP_CENTRAL_DIRECTORY_SIG)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "wrong zip central directory signature (0x%x)", sig);
+				fz_throw(ctx, FZ_ERROR_FORMAT, "wrong zip central directory signature (0x%x)", sig);
 
 			(void) fz_read_uint16_le(ctx, file); /* version made by */
 			(void) fz_read_uint16_le(ctx, file); /* version to extract */
@@ -355,7 +356,7 @@ static void read_zip_dir_imp(fz_context *ctx, fz_zip_archive *zip, int64_t start
 
 			n = fz_read(ctx, file, (unsigned char*)name, namesize);
 			if (n < (size_t)namesize)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "premature end of data in zip entry name");
+				fz_throw(ctx, FZ_ERROR_FORMAT, "premature end of data in zip entry name");
 			name[namesize] = '\0';
 
 			if (!utf8)
@@ -386,12 +387,29 @@ static void read_zip_dir_imp(fz_context *ctx, fz_zip_archive *zip, int64_t start
 					}
 					fz_seek(ctx, file, sizeleft - size, 1);
 				}
+				if (type == ZIP_UP_SIG && size > 5)
+				{
+					int sizeleft = size - 1;
+					if (fz_read_byte(ctx, file) == 1)
+					{
+						/* Version 1 */
+						(void) fz_read_uint32(ctx, file); /* Skip the CRC */
+						sizeleft -= 4;
+						fz_free(ctx, name);
+						name = NULL;
+						name = Memento_label(fz_malloc(ctx, sizeleft + 1), "zip_name");
+						fz_read(ctx, file, (unsigned char *)name, sizeleft);
+						name[sizeleft] = 0;
+						sizeleft = 0;
+					}
+					fz_seek(ctx, file, sizeleft - size, 1);
+				}
 				fz_seek(ctx, file, size, 1);
 				metasize -= 4 + size;
 			}
 
 			if (usize > INT32_MAX || csize > INT32_MAX)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "zip archive entry larger than 2 GB");
+				fz_throw(ctx, FZ_ERROR_FORMAT, "zip archive entry larger than 2 GB");
 
 			fz_seek(ctx, file, commentsize, 1);
 
@@ -422,12 +440,12 @@ static int read_zip_entry_header(fz_context *ctx, fz_zip_archive *zip, zip_entry
 
 	sig = fz_read_uint32_le(ctx, file);
 	if (sig != ZIP_LOCAL_FILE_SIG)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "wrong zip local file signature (0x%x)", sig);
+		fz_throw(ctx, FZ_ERROR_FORMAT, "wrong zip local file signature (0x%x)", sig);
 
 	(void) fz_read_uint16_le(ctx, file); /* version */
 	general = fz_read_uint16_le(ctx, file); /* general */
 	if (general & ZIP_ENCRYPTED_FLAG)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "zip content is encrypted");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "zip content is encrypted");
 
 	method = fz_read_uint16_le(ctx, file);
 	(void) fz_read_uint16_le(ctx, file); /* file time */
@@ -471,7 +489,7 @@ static void ensure_zip_entries(fz_context *ctx, fz_zip_archive *zip)
 		back += sizeof buf - 4;
 	}
 
-	fz_throw(ctx, FZ_ERROR_GENERIC, "cannot find end of central directory");
+	fz_throw(ctx, FZ_ERROR_FORMAT, "cannot find end of central directory");
 }
 
 static zip_entry *lookup_zip_entry(fz_context *ctx, fz_zip_archive *zip, const char *name)
@@ -501,7 +519,7 @@ static fz_stream *open_zip_entry(fz_context *ctx, fz_archive *arch, const char *
 		return fz_open_null_filter(ctx, file, ent->usize, fz_tell(ctx, file));
 	if (method == 8)
 		return fz_open_flated(ctx, file, -15);
-	fz_throw(ctx, FZ_ERROR_GENERIC, "unknown zip method: %d", method);
+	fz_throw(ctx, FZ_ERROR_FORMAT, "unknown zip method: %d", method);
 }
 
 static fz_buffer *read_zip_entry(fz_context *ctx, fz_archive *arch, const char *name)
@@ -546,31 +564,31 @@ static fz_buffer *read_zip_entry(fz_context *ctx, fz_archive *arch, const char *
 		{
 			cbuf = fz_malloc(ctx, ent->csize);
 
-			fz_read(ctx, file, cbuf, ent->csize);
-
 			z.zalloc = fz_zlib_alloc;
 			z.zfree = fz_zlib_free;
 			z.opaque = ctx;
-			z.next_in = cbuf;
-			z.avail_in = ent->csize;
 			z.next_out = ubuf->data;
 			z.avail_out = ent->usize;
+			z.next_in = cbuf;
+			z.avail_in = fz_read(ctx, file, cbuf, ent->csize);
+			if (z.avail_in < ent->csize)
+				fz_warn(ctx, "premature end of compressed data for compressed archive entry");
 
 			code = inflateInit2(&z, -15);
 			if (code != Z_OK)
 			{
-				fz_throw(ctx, FZ_ERROR_GENERIC, "zlib inflateInit2 error: %s", z.msg);
+				fz_throw(ctx, FZ_ERROR_LIBRARY, "zlib inflateInit2 error: %s", z.msg);
 			}
 			code = inflate(&z, Z_FINISH);
 			if (code != Z_STREAM_END)
 			{
 				inflateEnd(&z);
-				fz_throw(ctx, FZ_ERROR_GENERIC, "zlib inflate error: %s", z.msg);
+				fz_throw(ctx, FZ_ERROR_LIBRARY, "zlib inflate error: %s", z.msg);
 			}
 			code = inflateEnd(&z);
 			if (code != Z_OK)
 			{
-				fz_throw(ctx, FZ_ERROR_GENERIC, "zlib inflateEnd error: %s", z.msg);
+				fz_throw(ctx, FZ_ERROR_LIBRARY, "zlib inflateEnd error: %s", z.msg);
 			}
 
 			len = ent->usize - z.avail_out;
@@ -591,7 +609,7 @@ static fz_buffer *read_zip_entry(fz_context *ctx, fz_archive *arch, const char *
 	}
 
 	fz_drop_buffer(ctx, ubuf);
-	fz_throw(ctx, FZ_ERROR_GENERIC, "unknown zip method: %d", method);
+	fz_throw(ctx, FZ_ERROR_FORMAT, "unknown zip method: %d", method);
 }
 
 static int has_zip_entry(fz_context *ctx, fz_archive *arch, const char *name)
@@ -622,6 +640,9 @@ fz_is_zip_archive(fz_context *ctx, fz_stream *file)
 	unsigned char data[4];
 	size_t n;
 
+	if (file == NULL)
+		return 0;
+
 	fz_seek(ctx, file, 0, 0);
 	n = fz_read(ctx, file, data, nelem(data));
 	if (n != nelem(signature))
@@ -638,7 +659,7 @@ fz_open_zip_archive_with_stream(fz_context *ctx, fz_stream *file)
 	fz_zip_archive *zip;
 
 	if (!fz_is_zip_archive(ctx, file))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot recognize zip archive");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "cannot recognize zip archive");
 
 	zip = fz_new_derived_archive(ctx, file, fz_zip_archive);
 	zip->super.format = "zip";
