@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2023 Artifex Software, Inc.
+// Copyright (C) 2004-2024 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -385,6 +385,8 @@ static void epub_load_accelerator(fz_context *ctx, epub_document *doc, fz_stream
 				fz_free(ctx, acc->pages_in_chapter);
 			fz_free(ctx, acc);
 			/* Swallow the error and run unaccelerated */
+			fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+			fz_report_error(ctx);
 			make_new = 1;
 		}
 	}
@@ -481,7 +483,7 @@ epub_parse_chapter(fz_context *ctx, epub_document *doc, epub_chapter *ch)
 
 	buf = fz_read_archive_entry(ctx, zip, ch->path);
 	fz_try(ctx)
-		html = fz_parse_xhtml(ctx, doc->set, zip, base_uri, buf, fz_user_css(ctx));
+		html = fz_parse_html(ctx, doc->set, zip, base_uri, buf, fz_user_css(ctx), 1, 1, 0);
 	fz_always(ctx)
 		fz_drop_buffer(ctx, buf);
 	fz_catch(ctx)
@@ -844,7 +846,7 @@ epub_parse_header(fz_context *ctx, epub_document *doc)
 			/* Further abuse base_uri to hold a temporary name. */
 			const size_t z0 = sizeof("META-INF/encryption.xml")-1;
 			if (sizeof(base_uri) <= prefix_len + z0)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "Prefix too long in epub");
+				fz_throw(ctx, FZ_ERROR_FORMAT, "Prefix too long in epub");
 			strcpy(base_uri + prefix_len, "META-INF/encryption.xml");
 			if (fz_has_archive_entry(ctx, zip, base_uri))
 			{
@@ -865,7 +867,7 @@ epub_parse_header(fz_context *ctx, epub_document *doc)
 		rootfile = fz_xml_find_down(rootfiles, "rootfile");
 		full_path = fz_xml_att(rootfile, "full-path");
 		if (!full_path)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "cannot find root file in EPUB");
+			fz_throw(ctx, FZ_ERROR_FORMAT, "cannot find root file in EPUB");
 
 		fz_dirname(base_uri+prefix_len, full_path, sizeof(base_uri) - prefix_len);
 
@@ -917,6 +919,8 @@ epub_parse_header(fz_context *ctx, epub_document *doc)
 				fz_catch(ctx)
 				{
 					fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+					fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+					fz_report_error(ctx);
 					fz_warn(ctx, "ignoring chapter %s", s);
 				}
 			}
@@ -964,7 +968,7 @@ epub_output_accelerator(fz_context *ctx, fz_document *doc_, fz_output *out)
 	fz_try(ctx)
 	{
 		if (doc->accel == NULL)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "No accelerator data to write");
+			fz_throw(ctx, FZ_ERROR_ARGUMENT, "No accelerator data to write");
 
 		fz_write_int32_le(ctx, out, MAGIC_ACCELERATOR);
 		fz_write_int32_le(ctx, out, MAGIC_ACCEL_EPUB);
@@ -986,6 +990,8 @@ epub_output_accelerator(fz_context *ctx, fz_document *doc_, fz_output *out)
 		fz_rethrow(ctx);
 }
 
+/* Takes ownership of zip. Will always eventually drop it.
+ * Never takes ownership of accel. */
 static fz_document *
 epub_init(fz_context *ctx, fz_archive *zip, fz_stream *accel)
 {
@@ -1030,63 +1036,85 @@ epub_init(fz_context *ctx, fz_archive *zip, fz_stream *accel)
 }
 
 static fz_document *
-epub_open_accel_document_with_stream(fz_context *ctx, fz_stream *file, fz_stream *accel)
+epub_open_document(fz_context *ctx, const fz_document_handler *handler, fz_stream *file, fz_stream *accel, fz_archive *dir, void *state)
 {
-	return epub_init(ctx, fz_open_zip_archive_with_stream(ctx, file), accel);
-}
-
-static fz_document *
-epub_open_accel_document(fz_context *ctx, const char *filename, const char *accel)
-{
-	fz_stream *afile = NULL;
+	fz_stream *file2 = NULL;
 	fz_document *doc;
+	fz_archive *zip = NULL;
+
+	if (file == NULL)
+	{
+		/* Directory case: file == NULL and dir == the directory. */
+		if (fz_has_archive_entry(ctx, dir, "META-INF/container.xml"))
+			file2 = file = fz_open_archive_entry(ctx, dir, "META-INF/container.xml");
+		else
+			file2 = file = fz_open_archive_entry(ctx, dir, "META-INF\\container.xml");
+		if (file == NULL)
+			fz_throw(ctx, FZ_ERROR_FORMAT, "Not an epub file");
+		zip = fz_keep_archive(ctx, dir);
+	}
+	else
+	{
+		/* File case: file != NULL and dir can be ignored. */
+		zip = fz_open_archive_with_stream(ctx, file);
+	}
 
-	if (accel)
-		afile = fz_open_file(ctx, accel);
 
 	fz_try(ctx)
-	{
-		if (strstr(filename, "META-INF/container.xml") || strstr(filename, "META-INF\\container.xml"))
-		{
-			char dirname[2048], *p;
-			fz_strlcpy(dirname, filename, sizeof dirname);
-			p = strstr(dirname, "META-INF");
-			*p = 0;
-			if (!dirname[0])
-				fz_strlcpy(dirname, ".", sizeof dirname);
-			doc = epub_init(ctx, fz_open_directory(ctx, dirname), afile);
-		}
-		else
-			doc = epub_init(ctx, fz_open_zip_archive(ctx, filename), afile);
-	}
+		doc = epub_init(ctx, zip, file);
 	fz_always(ctx)
-		fz_drop_stream(ctx, afile);
+		fz_drop_stream(ctx, file2);
 	fz_catch(ctx)
 		fz_rethrow(ctx);
 
 	return doc;
 }
 
-static fz_document *
-epub_open_document_with_stream(fz_context *ctx, fz_stream *file)
-{
-	return epub_init(ctx, fz_open_zip_archive_with_stream(ctx, file), NULL);
-}
-
-static fz_document *
-epub_open_document(fz_context *ctx, const char *filename)
-{
-	return epub_open_accel_document(ctx, filename, NULL);
-}
-
 static int
-epub_recognize(fz_context *doc, const char *magic)
+epub_recognize(fz_context *doc, const fz_document_handler *handler, const char *magic)
 {
 	if (strstr(magic, "META-INF/container.xml") || strstr(magic, "META-INF\\container.xml"))
 		return 200;
 	return 0;
 }
 
+static int
+epub_recognize_content(fz_context *ctx, const fz_document_handler *handler, fz_stream *stream, fz_archive *dir, void **state, fz_document_recognize_state_free_fn **free_state)
+{
+	fz_archive *arch = NULL;
+	int ret = 0;
+
+	fz_var(arch);
+	fz_var(ret);
+
+	if (state)
+		*state = NULL;
+	if (free_state)
+		*free_state = NULL;
+
+	fz_try(ctx)
+	{
+		if (stream == NULL)
+			arch = fz_keep_archive(ctx, dir);
+		else
+		{
+			arch = fz_try_open_archive_with_stream(ctx, stream);
+			if (arch == NULL)
+				break;
+		}
+
+		if (fz_has_archive_entry(ctx, arch, "META-INF/container.xml") ||
+			fz_has_archive_entry(ctx, arch, "META-INF\\container.xml"))
+			ret = 100;
+	}
+	fz_always(ctx)
+		fz_drop_archive(ctx, arch);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+
+	return ret;
+}
+
 static const char *epub_extensions[] =
 {
 	"epub",
@@ -1103,9 +1131,7 @@ fz_document_handler epub_document_handler =
 {
 	epub_recognize,
 	epub_open_document,
-	epub_open_document_with_stream,
 	epub_extensions,
 	epub_mimetypes,
-	epub_open_accel_document,
-	epub_open_accel_document_with_stream
+	epub_recognize_content
 };
