@@ -50,7 +50,7 @@ typedef struct
 	pdf_obj *obj;
 } pdf_link;
 
-void pdf_load_annots(fz_context *ctx, pdf_page *page, pdf_obj *annots);
+void pdf_load_annots(fz_context *ctx, pdf_page *page);
 void pdf_drop_annots(fz_context *ctx, pdf_annot *annot_list);
 void pdf_drop_widgets(fz_context *ctx, pdf_annot *widget_list);
 
@@ -64,4 +64,7 @@ pdf_obj *pdf_add_url_filespec(fz_context *ctx, pdf_document *doc, const char *ur
 char *pdf_parse_link_dest(fz_context *ctx, pdf_document *doc, pdf_obj *dest);
 char *pdf_parse_link_action(fz_context *ctx, pdf_document *doc, pdf_obj *obj, int pagenum);
 
+/* Rect diff is not really a rect! Just returned as such for convenience. */
+fz_rect pdf_annot_rect_diff(fz_context *ctx, pdf_annot *annot);
+
 #endif
