@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2023 Artifex Software, Inc.
+// Copyright (C) 2004-2024 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -156,7 +156,9 @@ pdf_load_image_imp(fz_context *ctx, pdf_document *doc, pdf_obj *rdb, pdf_obj *di
 				decode[i] = i & 1 ? maxval : 0;
 		}
 
-		obj = pdf_dict_geta(ctx, dict, PDF_NAME(SMask), PDF_NAME(Mask));
+		obj = pdf_dict_get(ctx, dict, PDF_NAME(SMask));
+		if (!pdf_is_dict(ctx, obj))
+			obj = pdf_dict_get(ctx, dict, PDF_NAME(Mask));
 		if (pdf_is_dict(ctx, obj))
 		{
 			/* Not allowed for inline images or soft masks */
@@ -332,7 +334,7 @@ struct jbig2_segment_header {
 	int flags;
 	/* referred-to-segment numbers */
 	int page;
-	int length;
+	uint32_t length;
 };
 
 /* coverity[-tainted_data_return] */
@@ -578,8 +580,8 @@ pdf_add_image(fz_context *ctx, pdf_document *doc, fz_image *image)
 				if (cp->u.jbig2.embedded && cp->u.jbig2.globals)
 				{
 					pdf_obj *globals_ref = pdf_add_new_dict(ctx, doc, 1);
+					pdf_dict_put_drop(ctx, dp, PDF_NAME(JBIG2Globals), globals_ref);
 					pdf_update_stream(ctx, doc, globals_ref, fz_jbig2_globals_data(ctx, cp->u.jbig2.globals), 0);
-					pdf_dict_put(ctx, dp, PDF_NAME(JBIG2Globals), globals_ref);
 				}
 				else
 					buffer = pdf_jbig2_stream_from_file(ctx, cbuffer->buffer,
