@@ -73,7 +73,7 @@ fz_recolor_shade_type1(fz_context *ctx, pdf_obj *shade, pdf_function **func, rec
 
 	if (rd->funcs != 1 && rd->funcs != n_in)
 	{
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Unexpected function-arity.");
+		fz_throw(ctx, FZ_ERROR_SYNTAX, "Unexpected function-arity.");
 	}
 
 	/* Sample the function, rewriting it. */
@@ -171,7 +171,7 @@ fz_recolor_shade_type1(fz_context *ctx, pdf_obj *shade, pdf_function **func, rec
 }
 
 static void
-fz_recolor_shade_function(fz_context *ctx, pdf_obj *shade, float samples[256][FZ_MAX_COLORS+1], recolor_details *rd)
+fz_recolor_shade_function(fz_context *ctx, pdf_obj *shade, float *samples, int stride, recolor_details *rd)
 {
 	int i;
 	int n_in = fz_colorspace_n(ctx, rd->src_cs);
@@ -205,7 +205,7 @@ fz_recolor_shade_function(fz_context *ctx, pdf_obj *shade, float samples[256][FZ
 	for (t = 0; t < 256; t++)
 	{
 		for (i = 0; i < n_in; i++)
-			p[i] = samples[t][i];
+			p[i] = samples[t*stride+i];
 
 		rd->recolor(ctx, rd->opaque, rd->dst_cs, q, rd->src_cs, p);
 
@@ -404,6 +404,7 @@ fz_recolor_shade_type4(fz_context *ctx, pdf_obj *shade, recolor_details *rd)
 
 	fz_var(outbuf);
 	fz_var(out);
+	fz_var(stream);
 
 	read_decode(ctx, shade, n_in, c_min, c_max, n_out, d_min, d_max);
 
@@ -497,6 +498,7 @@ fz_recolor_shade_type5(fz_context *ctx, pdf_obj *shade, recolor_details *rd)
 
 	fz_var(outbuf);
 	fz_var(out);
+	fz_var(stream);
 
 	read_decode(ctx, shade, n_in, c_min, c_max, n_out, d_min, d_max);
 
@@ -592,6 +594,7 @@ fz_recolor_shade_type6(fz_context *ctx, pdf_obj *shade, recolor_details *rd)
 
 	fz_var(outbuf);
 	fz_var(out);
+	fz_var(stream);
 
 	read_decode(ctx, shade, n_in, c_min, c_max, n_out, d_min, d_max);
 
@@ -731,6 +734,7 @@ fz_recolor_shade_type7(fz_context *ctx, pdf_obj *shade, recolor_details *rd)
 
 	fz_var(outbuf);
 	fz_var(out);
+	fz_var(stream);
 
 	read_decode(ctx, shade, n_in, c_min, c_max, n_out, d_min, d_max);
 
@@ -859,7 +863,7 @@ pdf_new_colorspace(fz_context *ctx, fz_colorspace *cs)
 	case FZ_COLORSPACE_CMYK:
 		return PDF_NAME(DeviceCMYK);
 	default:
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Unimplemented colorspace");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Unimplemented colorspace");
 	}
 
 	return NULL;
@@ -877,7 +881,7 @@ pdf_recolor_shade(fz_context *ctx, pdf_obj *shade, pdf_shade_recolorer *reshade,
 	int type, i;
 	pdf_function *func[FZ_MAX_COLORS] = { NULL };
 	float d0, d1;
-	float samples[256][FZ_MAX_COLORS + 1];
+	float samples[256*(FZ_MAX_COLORS + 1)];
 	pdf_document *doc = pdf_get_bound_document(ctx, shade);
 
 	src_cs = pdf_load_colorspace(ctx, pdf_dict_get(ctx, shade, PDF_NAME(ColorSpace)));
@@ -909,9 +913,9 @@ pdf_recolor_shade(fz_context *ctx, pdf_obj *shade, pdf_shade_recolorer *reshade,
 			float nbg[FZ_MAX_COLORS];
 
 			if (n > FZ_MAX_COLORS)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "Too many background components");
+				fz_throw(ctx, FZ_ERROR_SYNTAX, "Too many background components");
 			if (n != src_cs->n)
-				fz_throw(ctx, FZ_ERROR_GENERIC, "Wrong background dimension");
+				fz_throw(ctx, FZ_ERROR_SYNTAX, "Wrong background dimension");
 
 			for (i = 0; i < n; i++)
 				bg[i] = pdf_array_get_real(ctx, background, i);
@@ -920,7 +924,7 @@ pdf_recolor_shade(fz_context *ctx, pdf_obj *shade, pdf_shade_recolorer *reshade,
 
 			new_bg = pdf_dict_put_array(ctx, rewritten, PDF_NAME(Background), rd.dst_cs->n);
 			for (i = 0; i < n; i++)
-				pdf_array_put(ctx, new_bg, i, pdf_new_real(ctx, bg[i]));
+				pdf_array_put_real(ctx, new_bg, i, bg[i]);
 			pdf_dict_put(ctx, rewritten, PDF_NAME(Background), new_bg);
 		}
 
@@ -985,7 +989,7 @@ pdf_recolor_shade(fz_context *ctx, pdf_obj *shade, pdf_shade_recolorer *reshade,
 		/* For all other function based shadings, we just rewrite the 1d function. */
 		if (rd.funcs)
 		{
-			fz_recolor_shade_function(ctx, rewritten, samples, &rd);
+			fz_recolor_shade_function(ctx, rewritten, samples, src_cs->n+1, &rd);
 			break;
 		}
 
@@ -1002,7 +1006,7 @@ pdf_recolor_shade(fz_context *ctx, pdf_obj *shade, pdf_shade_recolorer *reshade,
 			break;
 		case FZ_LINEAR:
 		case FZ_RADIAL:
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Linear/Radial shadings must use functions");
+			fz_throw(ctx, FZ_ERROR_SYNTAX, "Linear/Radial shadings must use functions");
 			break;
 		case FZ_MESH_TYPE4:
 			fz_recolor_shade_type4(ctx, rewritten, &rd);
@@ -1017,7 +1021,7 @@ pdf_recolor_shade(fz_context *ctx, pdf_obj *shade, pdf_shade_recolorer *reshade,
 			fz_recolor_shade_type7(ctx, rewritten, &rd);
 			break;
 		default:
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Unexpected mesh type %d\n", type);
+			fz_throw(ctx, FZ_ERROR_SYNTAX, "Unexpected mesh type %d\n", type);
 		}
 	}
 	fz_always(ctx)
