@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2024 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -75,7 +75,7 @@ svg_load_page(fz_context *ctx, fz_document *doc_, int chapter, int number)
 	svg_page *page;
 
 	if (number != 0)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot find page %d", number);
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "cannot find page %d", number);
 
 	page = fz_new_derived_page(ctx, svg_page, doc_);
 	page->super.bound_page = svg_bound_page;
@@ -163,12 +163,11 @@ svg_open_document_with_buffer(fz_context *ctx, fz_buffer *buf, const char *base_
 }
 
 static fz_document *
-svg_open_document_with_stream(fz_context *ctx, fz_stream *file)
+svg_open_document(fz_context *ctx, const fz_document_handler *handler, fz_stream *file, fz_stream *accel, fz_archive *zip, void *state)
 {
-	fz_buffer *buf;
+	fz_buffer *buf = fz_read_all(ctx, file, 0);
 	fz_document *doc = NULL;
 
-	buf = fz_read_all(ctx, file, 0);
 	fz_try(ctx)
 		doc = svg_open_document_with_buffer(ctx, buf, NULL, NULL);
 	fz_always(ctx)
@@ -268,59 +267,67 @@ static const char *svg_mimetypes[] =
 };
 
 static int
-svg_recognize_doc_content(fz_context *ctx, fz_stream *stream)
+svg_recognize_doc_content(fz_context *ctx, const fz_document_handler *handler, fz_stream *stm, fz_archive *dir, void **state, fz_document_recognize_state_free_fn **free_state)
 {
+	// A standalone SVG document is an XML document with an <svg> root element.
+	//
+	// Assume the document is ASCII or UTF-8.
+	//
+	// Parse the start of the file using a simplified XML parser, skipping
+	// processing instructions and comments, and stopping at the first
+	// element.
+	//
+	// Return failure on anything unexpected, or if the first element is not SVG.
+
 	int c;
-	int n = 0;
-	const char *match = "svg";
-	int pos = 0;
 
-	/* Is the first non-whitespace char '<' ? */
-	do
-	{
-		c = fz_read_byte(ctx, stream);
-		if (c == EOF)
-			return 0;
-		if (c == '<')
-			break;
-		if (c != ' ' && c != '\t' && c != '\n' && c != '\r')
-			return 0;
-	}
-	while (++n < 4096);
+	if (state)
+		*state = NULL;
+	if (free_state)
+		*free_state = NULL;
 
-	/* Then do we find 'svg' in the first 4k? */
-	do
-	{
-		c = fz_read_byte(ctx, stream);
-		if (c == EOF)
-			return 0;
-		if (c >= 'A' && c <= 'Z')
-			c += 'a' - 'A';
-		if (c == match[pos])
-		{
-			pos++;
-			if (pos == 3)
-				return 100;
-		}
-		else
-		{
-			/* Restart matching, but recheck c against the start. */
-			pos = (c == match[0]);
-		}
-	}
-	while (++n < 4096);
+	if (stm == NULL)
+		return 0;
 
+parse_text:
+	// Skip whitespace until "<"
+	c = fz_read_byte(ctx, stm);
+	while (c == ' ' || c == '\r' || c == '\n' || c == '\t')
+		c = fz_read_byte(ctx, stm);
+	if (c == '<')
+		goto parse_element;
+	return 0;
+
+parse_element:
+	// Either "<?...>" or "<!...>" or "<svg" or not an SVG document.
+	c = fz_read_byte(ctx, stm);
+	if (c == '!' || c == '?')
+		goto parse_comment;
+	if (c != 's')
+		return 0;
+	c = fz_read_byte(ctx, stm);
+	if (c != 'v')
+		return 0;
+	c = fz_read_byte(ctx, stm);
+	if (c != 'g')
+		return 0;
+	return 100;
+
+parse_comment:
+	// Skip everything after "<?" or "<!" until ">"
+	c = fz_read_byte(ctx, stm);
+	while (c != EOF && c != '>')
+		c = fz_read_byte(ctx, stm);
+	if (c == '>')
+		goto parse_text;
 	return 0;
 }
 
 fz_document_handler svg_document_handler =
 {
 	NULL,
-	NULL,
-	svg_open_document_with_stream,
+	svg_open_document,
 	svg_extensions,
 	svg_mimetypes,
-	NULL,
-	NULL,
 	svg_recognize_doc_content
 };
