@@ -650,7 +650,7 @@ int pdfshow_main(int argc, char **argv)
 	}
 	fz_catch(ctx)
 	{
-		fz_log_error(ctx, fz_caught_message(ctx));
+		fz_report_error(ctx);
 		errored = 1;
 	}
 
