@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 #include "xps-imp.h"
@@ -59,7 +59,7 @@ xps_parse_document_outline(fz_context *ctx, xps_document *doc, fz_xml *root)
 			entry = fz_new_outline(ctx);
 			entry->title = Memento_label(fz_strdup(ctx, description), "outline_title");
 			entry->uri = Memento_label(fz_strdup(ctx, target), "outline_uri");
-			entry->page = xps_lookup_link_target(ctx, (fz_document*)doc, target, NULL, NULL).page;
+			entry->page = xps_lookup_link_target(ctx, (fz_document*)doc, target).loc;
 			entry->down = NULL;
 			entry->next = NULL;
 
@@ -147,6 +147,8 @@ xps_load_outline(fz_context *ctx, fz_document *doc_)
 			fz_catch(ctx)
 			{
 				fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+				fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+				fz_report_error(ctx);
 				outline = NULL;
 			}
 			if (!outline)
